#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating pure component properties from:

Perry's Chemical Engineers' Handbook, 7th Edition
Perry, Green, Maloney, 1997, McGraw-Hill

All parameter indices and units based on conventions used by the source
"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import log, Var, Param, units as pyunits

from idaes.core.util.misc import set_param_from_config

from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


# -----------------------------------------------------------------------------
class Perrys(object):
    # -----------------------------------------------------------------------------
    # Heat capacities, enthalpies and entropies
    class cp_mol_liq_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.cp_mol_liq_comp_coeff_1 = Var(
                doc="Parameter 1 for liquid phase molar heat capacity",
                units=pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="1")

            cobj.cp_mol_liq_comp_coeff_2 = Var(
                doc="Parameter 2 for liquid phase molar heat capacity",
                units=pyunits.J * pyunits.kmol**-1 * pyunits.K**-2,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="2")

            cobj.cp_mol_liq_comp_coeff_3 = Var(
                doc="Parameter 3 for liquid phase molar heat capacity",
                units=pyunits.J * pyunits.kmol**-1 * pyunits.K**-3,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="3")

            cobj.cp_mol_liq_comp_coeff_4 = Var(
                doc="Parameter 4 for liquid phase molar heat capacity",
                units=pyunits.J * pyunits.kmol**-1 * pyunits.K**-4,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="4")

            cobj.cp_mol_liq_comp_coeff_5 = Var(
                doc="Parameter 5 for liquid phase molar heat capacity",
                units=pyunits.J * pyunits.kmol**-1 * pyunits.K**-5,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="5")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            T = pyunits.convert(T, to_units=pyunits.K)
            cp = (
                cobj.cp_mol_liq_comp_coeff_5 * T**4
                + cobj.cp_mol_liq_comp_coeff_4 * T**3
                + cobj.cp_mol_liq_comp_coeff_3 * T**2
                + cobj.cp_mol_liq_comp_coeff_2 * T
                + cobj.cp_mol_liq_comp_coeff_1
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(cp, units.HEAT_CAPACITY_MOLE)

    class enth_mol_liq_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_liq_comp_coeff_1"):
                Perrys.cp_mol_liq_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_liq_comp_ref = Var(
                    doc="Liquid phase molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_liq_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            h_form = (
                cobj.enth_mol_form_liq_comp_ref
                if b.params.config.include_enthalpy_of_formation
                else 0 * units.ENERGY_MOLE
            )

            h = (
                pyunits.convert(
                    (cobj.cp_mol_liq_comp_coeff_5 / 5) * (T**5 - Tr**5)
                    + (cobj.cp_mol_liq_comp_coeff_4 / 4) * (T**4 - Tr**4)
                    + (cobj.cp_mol_liq_comp_coeff_3 / 3) * (T**3 - Tr**3)
                    + (cobj.cp_mol_liq_comp_coeff_2 / 2) * (T**2 - Tr**2)
                    + cobj.cp_mol_liq_comp_coeff_1 * (T - Tr),
                    units.ENERGY_MOLE,
                )
                + h_form
            )

            return h

    class entr_mol_liq_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_liq_comp_coeff_1"):
                Perrys.cp_mol_liq_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_liq_comp_ref = Var(
                doc="Liquid phase molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_liq_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            s = (
                pyunits.convert(
                    (cobj.cp_mol_liq_comp_coeff_5 / 4) * (T**4 - Tr**4)
                    + (cobj.cp_mol_liq_comp_coeff_4 / 3) * (T**3 - Tr**3)
                    + (cobj.cp_mol_liq_comp_coeff_3 / 2) * (T**2 - Tr**2)
                    + cobj.cp_mol_liq_comp_coeff_2 * (T - Tr)
                    + cobj.cp_mol_liq_comp_coeff_1 * log(T / Tr),
                    units.ENTROPY_MOLE,
                )
                + cobj.entr_mol_form_liq_comp_ref
            )

            return s

    # -----------------------------------------------------------------------------
    # Densities

    class dens_mol_liq_comp_eqn_1:
        @staticmethod
        def build_parameters(cobj):
            cobj.dens_mol_liq_comp_coeff_1 = Var(
                doc="Parameter 1 for liquid phase molar density",
                units=pyunits.kmol * pyunits.m**-3,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="1")

            cobj.dens_mol_liq_comp_coeff_2 = Var(
                doc="Parameter 2 for liquid phase molar density",
                units=pyunits.dimensionless,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="2")

            cobj.dens_mol_liq_comp_coeff_3 = Var(
                doc="Parameter 3 for liquid phase molar density", units=pyunits.K
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="3")

            cobj.dens_mol_liq_comp_coeff_4 = Var(
                doc="Parameter 4 for liquid phase molar density",
                units=pyunits.dimensionless,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="4")

        @staticmethod
        def return_expression(b, cobj, T):
            # pg. 2-98
            T = pyunits.convert(T, to_units=pyunits.K)

            rho = cobj.dens_mol_liq_comp_coeff_1 / cobj.dens_mol_liq_comp_coeff_2 ** (
                1
                + (1 - T / cobj.dens_mol_liq_comp_coeff_3)
                ** cobj.dens_mol_liq_comp_coeff_4
            )

            units = b.params.get_metadata().derived_units

            return pyunits.convert(rho, units.DENSITY_MOLE)

    class dens_mol_liq_comp_eqn_2:
        @staticmethod
        def build_parameters(cobj):
            cobj.dens_mol_liq_comp_coeff_1 = Var(
                doc="Parameter 1 for liquid phase molar density",
                units=pyunits.kmol / pyunits.m**3,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="1")

            cobj.dens_mol_liq_comp_coeff_2 = Var(
                doc="Parameter 2 for liquid phase molar density",
                units=pyunits.kmol / pyunits.m**3 / pyunits.K,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="2")

            cobj.dens_mol_liq_comp_coeff_3 = Var(
                doc="Parameter 3 for liquid phase molar density",
                units=pyunits.kmol / pyunits.m**3 / pyunits.K**2,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="3")

            cobj.dens_mol_liq_comp_coeff_4 = Var(
                doc="Parameter 4 for liquid phase molar density",
                units=pyunits.kmol / pyunits.m**3 / pyunits.K**3,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="4")

        @staticmethod
        def return_expression(b, cobj, T):
            # pg. 2-98
            T = pyunits.convert(T, to_units=pyunits.K)

            rho = (
                (cobj.dens_mol_liq_comp_coeff_1)
                + (cobj.dens_mol_liq_comp_coeff_2) * T
                + (cobj.dens_mol_liq_comp_coeff_3) * T**2
                + (cobj.dens_mol_liq_comp_coeff_4) * T**3
            )

            units = b.params.get_metadata().derived_units

            return pyunits.convert(rho, units.DENSITY_MOLE)

    class dens_mol_liq_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.dens_mol_liq_comp_coeff_eqn_type = Param(
                mutable=True, doc="Liquid molar density equation form"
            )

            set_param_from_config(
                cobj, param="dens_mol_liq_comp_coeff", index="eqn_type"
            )

            if cobj.dens_mol_liq_comp_coeff_eqn_type.value == 1:
                Perrys.dens_mol_liq_comp_eqn_1.build_parameters(cobj)
            elif cobj.dens_mol_liq_comp_coeff_eqn_type.value == 2:
                Perrys.dens_mol_liq_comp_eqn_2.build_parameters(cobj)
            else:
                raise ConfigurationError(
                    f"{cobj.name} unrecognized value for "
                    f"dens_mol_liq_comp equation type: "
                    f"{cobj.dens_mol_liq_comp_coeff_eqn_type}"
                )

        @staticmethod
        def return_expression(b, cobj, T):
            if cobj.dens_mol_liq_comp_coeff_eqn_type.value == 1:
                rho = Perrys.dens_mol_liq_comp_eqn_1.return_expression(b, cobj, T)
            elif cobj.dens_mol_liq_comp_coeff_eqn_type.value == 2:
                rho = Perrys.dens_mol_liq_comp_eqn_2().return_expression(b, cobj, T)
            else:
                raise ConfigurationError(
                    "No expression for eqn_type of "
                    "dens_mol_liq_comp_coeff specified,"
                    "please specify valid flag."
                )
            return rho



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for surrogates/sampling/scaling module
"""
import pytest
import pandas as pd
from idaes.core.surrogate.sampling.scaling import OffsetScaler


class TestScaling:
    @pytest.mark.unit
    def test_OffsetScaler(self):
        df = pd.DataFrame(
            {"A": [1.0, 2.0, 3.0, 4.0, 5.0], "B": [2.0, 5.0, 6.0, 10.0, 12.0]}
        )

        scaler = OffsetScaler.create_from_mean_std(df)
        pd.testing.assert_series_equal(scaler._offset, pd.Series({"A": 3.0, "B": 7.0}))
        pd.testing.assert_series_equal(
            scaler._factor, pd.Series({"A": 1.58113, "B": 4.0})
        )

        scaled_df = scaler.scale(df)
        expected_scaled_df = pd.DataFrame(
            {
                "A": [-1.264911064, -0.632455532, 0, 0.632455532, 1.264911064],
                "B": [-1.25, -0.5, -0.25, 0.75, 1.25],
            }
        )
        pd.testing.assert_frame_equal(scaled_df, expected_scaled_df)

        unscaled_df = scaler.unscale(scaled_df)
        pd.testing.assert_frame_equal(df, unscaled_df)

        # test that error is thrown if columns do not match
        with pytest.raises(ValueError):
            df2 = pd.DataFrame({"A": [1, 2, 3, 4, 5], "B2": [2, 5, 6, 10, 12]})
            scaled_df2 = scaler.scale(df2)

        with pytest.raises(ValueError) as excinfo:
            scaler = OffsetScaler(
                expected_columns=["A", "B"],
                offset_series=pd.Series({"A": 1, "C": 2}),
                factor_series=pd.Series({"A": 3, "B": 4}),
            )
        assert (
            str(excinfo.value)
            == "OffsetScaler was passed an offset series with an index that"
            " does not match expected_columns. Please make sure these labels match."
        )

        with pytest.raises(ValueError) as excinfo:
            scaler = OffsetScaler(
                expected_columns=["A", "B"],
                offset_series=pd.Series({"A": 1, "B": 2}),
                factor_series=pd.Series({"A": 3, "C": 4}),
            )
        assert (
            str(excinfo.value)
            == "OffsetScaler was passed a factor series with an index that"
            " does not match expected_columns. Please make sure these labels match."
        )

    @pytest.mark.unit
    def test_OffsetScaler_normalize(self):
        df = pd.DataFrame(
            {"A": [1.0, 2.0, 3.0, 4.0, 5.0], "B": [2.0, 5.0, 6.0, 10.0, 12.0]}
        )

        scaler = OffsetScaler.create_normalizing_scaler(df)
        pd.testing.assert_series_equal(scaler._offset, pd.Series({"A": 1.0, "B": 2.0}))
        pd.testing.assert_series_equal(scaler._factor, pd.Series({"A": 4.0, "B": 10.0}))

        scaled_df = scaler.scale(df)
        expected_scaled_df = pd.DataFrame(
            {"A": [0.0, 0.25, 0.5, 0.75, 1.0], "B": [0.0, 0.3, 0.4, 0.8, 1.0]}
        )
        pd.testing.assert_frame_equal(scaled_df, expected_scaled_df)

        df = pd.DataFrame({"A": [0.0, 6.0], "B": [3.0, 7.0]})
        scaled_df = scaler.scale(df)
        expected_scaled_df = pd.DataFrame({"A": [-0.25, 1.25], "B": [0.1, 0.5]})
        pd.testing.assert_frame_equal(scaled_df, expected_scaled_df)

        unscaled_df = scaler.unscale(scaled_df)
        pd.testing.assert_frame_equal(df, unscaled_df)

    @pytest.mark.unit
    def test_OffsetScaler_serialization(self):
        df = pd.DataFrame(
            {"A": [4.0, 10.0, 12.0, 20.0, 24.0], "B": [2.0, 5.0, 6.0, 10.0, 12.0]}
        )
        scaler = OffsetScaler.create_from_mean_std(df)

        d = scaler.to_dict()
        assert d["expected_columns"] == ["A", "B"]
        assert d["offset"]["A"] == pytest.approx(14)
        assert d["offset"]["B"] == pytest.approx(7)
        assert d["factor"]["A"] == pytest.approx(8)
        assert d["factor"]["B"] == pytest.approx(4)

        scaler = OffsetScaler.from_dict(d)
        assert scaler._expected_columns == ["A", "B"]
        pd.testing.assert_series_equal(scaler._offset, pd.Series({"A": 14.0, "B": 7.0}))
        pd.testing.assert_series_equal(scaler._factor, pd.Series({"A": 8.0, "B": 4.0}))


if __name__ == "__main__":
    pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for Scalers

Author: Andrew Lee
"""

from pyomo.common.config import (
    Bool,
    ConfigDict,
    ConfigValue,
    String_ConfigFormatter,
)
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData

from idaes.core.scaling.util import get_scaling_factor, set_scaling_factor
import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)


# Common base ConfigBlock for all Scalers
CONFIG = ConfigDict()
CONFIG.declare(
    "zero_tolerance",
    ConfigValue(
        default=1e-12,
        domain=float,
        description="Value at which a variable will be considered equal to zero for scaling.",
    ),
)
CONFIG.declare(
    "max_variable_scaling_factor",
    ConfigValue(
        default=1e10,
        domain=float,
        description="Maximum value for variable scaling factors.",
    ),
)
CONFIG.declare(
    "min_variable_scaling_factor",
    ConfigValue(
        default=1e-10,
        domain=float,
        description="Minimum value for variable scaling factors.",
    ),
)
CONFIG.declare(
    "max_constraint_scaling_factor",
    ConfigValue(
        default=1e10,
        domain=float,
        description="Maximum value for constraint scaling factors.",
    ),
)
CONFIG.declare(
    "min_constraint_scaling_factor",
    ConfigValue(
        default=1e-10,
        domain=float,
        description="Minimum value for constraint scaling factors.",
    ),
)
CONFIG.declare(
    "overwrite",
    ConfigValue(
        default=False,
        domain=Bool,
        description="Whether to overwrite existing scaling factors.",
    ),
)


class ScalerBase:
    """
    Base class for IDAES Scaler objects

    Contains a number of methods useful for scaling models.

    """

    CONFIG = CONFIG()

    def __init__(self, **kwargs):
        self.config = self.CONFIG(kwargs)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Handle cases where a class doc string was not set
        if cls.__doc__ is None:
            cls.__doc__ = ""

        cls.__doc__ = cls.__doc__ + cls.CONFIG.generate_documentation(
            format=String_ConfigFormatter(
                block_start="%s\n",
                block_end="",
                item_start="%s\n",
                item_body="%s",
                item_end="\n",
            ),
            indent_spacing=4,
            width=66,
        )

    def get_scaling_factor(self, component):
        """
        Get scaling factor for component.

        Alias for get_scaling_factor utility function.

        Args:
            component: component to get scaling factor for

        Returns:
            float - scaling factor

        Raises:
            TypeError if component is a Block
        """
        return get_scaling_factor(component)

    def set_variable_scaling_factor(
        self, variable, scaling_factor: float, overwrite: bool = None
    ):
        """
        Set scaling factor for variable.

        Scaling factor is limited by min_variable_scaling_factor and max_variable_scaling_factor.

        Args:
            variable: VarData component to set scaling factor for.
            scaling_factor: nominal scaling factor to apply. May be limited by max and min values.
            overwrite: whether to overwrite existing scaling factor (if present).
              Defaults to Scaler config setting.

        Returns:
            None

        Raises:
            TypeError if variable is not an instance of VarData
        """
        if not isinstance(variable, VarData):
            raise TypeError(f"{variable} is not a variable (or is indexed).")
        self._set_scaling_factor(
            component=variable,
            component_type="variable",
            scaling_factor=scaling_factor,
            overwrite=overwrite,
        )

    def set_constraint_scaling_factor(
        self, constraint, scaling_factor: float, overwrite: bool = None
    ):
        """
        Set scaling factor for constraint.

        Scaling factor is limited by min_constraint_scaling_factor and max_constraint_scaling_factor.

        Args:
            constraint: ConstraintData component to set scaling factor for.
            scaling_factor: nominal scaling factor to apply. May be limited by max and min values.
            overwrite: whether to overwrite existing scaling factor (if present).
              Defaults to Scaler config setting.

        Returns:
            None

        Raises:
            TypeError if constraint is not an instance of ConstraintData
        """
        if not isinstance(constraint, ConstraintData):
            raise TypeError(f"{constraint} is not a constraint (or is indexed).")
        self._set_scaling_factor(
            component=constraint,
            component_type="constraint",
            scaling_factor=scaling_factor,
            overwrite=overwrite,
        )

    def _set_scaling_factor(
        self, component, component_type, scaling_factor, overwrite=None
    ):
        """
        PRIVATE METHOD

        The purpose of this method is to apply the correct max and min limits to scaling factors
        before setting them. Which set of limits to apply is determined solely by the component_type
        argument, and no type checking is performed.
        """
        # If overwrite not provided, use scaler config
        # This allows developers to have some more control over when to overwrite or not
        if overwrite is None:
            overwrite = self.config.overwrite

        if component_type == "variable":
            maxsf = self.config.max_variable_scaling_factor
            minsf = self.config.min_variable_scaling_factor
        elif component_type == "constraint":
            maxsf = self.config.max_constraint_scaling_factor
            minsf = self.config.min_constraint_scaling_factor
        else:
            raise ValueError("Invalid value for component_type.")

        if scaling_factor > maxsf:
            _log.debug(
                f"Scaling factor for {component.name} limited by maximum value "
                f"(max_sf: {maxsf} < sf: {scaling_factor})"
            )
            scaling_factor = maxsf
        elif scaling_factor < minsf:
            _log.debug(
                f"Scaling factor for {component.name} limited by minimum value "
                f"(min_sf: {minsf} > sf: {scaling_factor})"
            )
            scaling_factor = minsf

        set_scaling_factor(component, scaling_factor, overwrite=overwrite)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

import math
import numpy as np


def split_training_validation(dataframe, training_fraction, seed=None):
    """
    Randomly split the dataframe into training and validation data

    Args:
       dataframe : pandas DataFrame
          The dataframe to split
       training_fraction : float between 0 < 1
          The fraction of the overall dataframe (# rows) to include
          as training data. The rest will be returned as validation data
       seed : None or int
          seed used for the random number generator. Use default behavior
          for DataFrame.sample if seed is None

    Returns:
       tuple : (training_dataframe, validation_dataframe)
    """
    return split_dataframe(dataframe, [training_fraction], seed)


def split_training_validation_testing(
    dataframe, training_fraction, validation_fraction, seed=None
):
    """
    Randomly split the dataframe into training, validation, and testing data

    Args:
       dataframe : pandas DataFrame
          The dataframe to split
       training_fraction : float between 0 < 1
          The fraction of the overall dataframe (# rows) to include
          as training data.
          The sum of training_fraction and validation_fraction must be less than 1
       validation_fraction : float between 0 < 1
          The fraction of the overall dataframe (# rows) to include
          as validation data. The rest will be returned as validation data.
          The sum of training_fraction and validation_fraction must be less than 1
       seed : None or int
          seed used for the random number generator. Use default behavior
          for DataFrame.sample if seed is None

    Returns:
       tuple : (training_dataframe, validation_dataframe, testing_dataframe)
    """
    return split_dataframe(dataframe, [training_fraction, validation_fraction], seed)


def split_dataframe(dataframe, fractions, seed=None):
    """
    Randomly splits the dataframe into multiple dataframes.

    Args:
       dataframe: pandas DataFrame
          The dataframe to split
       fractions: list of floats between 0 < 1
          The fraction of the data to include in each dataframe. The list of fractions
          must sum to < 1. If fractions has length N, then N+1 dataframes will be returned
          where the fraction for the last dataframe is 1-sum(fractions).
       seed : None or int
          seed for the random number generator. If None, generator is not seeded.

    Returns:
       tuple : (DataFrame, ...)
    """
    assert sum(fractions) < 1.0

    # note seed=None is the default value for random_state (e.g., not seeded)
    shuffled_df = dataframe.sample(frac=1, random_state=seed).reset_index(drop=True)

    dfs = np.split(
        shuffled_df, [math.floor(f * len(shuffled_df)) for f in np.cumsum(fractions)]
    )
    # reset all the indices
    for df in dfs:
        df.reset_index(drop=True, inplace=True)
    return dfs



# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains custom IDAES exceptions.
"""

__author__ = "Andrew Lee"


class IdaesError(Exception):
    """
    General exception for all IDAES errors. Intended to be used for multiple
    inheritance in derived exceptions to allow catching of all IDAES-related
    exceptions.
    """

    # Problem with toaster


class BalanceTypeNotSupportedError(NotImplementedError, IdaesError):
    """
    IDAES exception to be used when a control volume does not support a given
    type of balance equation.
    """

    # Tried to put bagel in normal toaster


class ConfigurationError(ValueError, IdaesError):
    """
    IDAES exception to be used when configuration arguments are incorrect
    or inconsistent.
    """

    # Too many buttons, burnt toast


class DynamicError(ValueError, IdaesError):
    """
    IDAES exception for cases where settings associated with dynamic models
    are incorrect.
    """

    # Incorrect browness setting


class BurntToast(IdaesError):
    """
    General exception for when something breaks badly in the core.
    """

    # Toaster on fire


class PropertyNotSupportedError(AttributeError, IdaesError):
    """
    IDAES exception for cases when a models calls for a property which is
    not supported by the chosen property package.

    Needs to inherit from AttributeError for Pyomo interactions.
    """

    # Could not find bread


class PropertyPackageError(AttributeError, IdaesError):
    """
    IDAES exception for generic errors arising from property packages.

    Needs to inherit from AttributeError for Pyomo interactions.
    """

    # Bread stuck


class InitializationError(ArithmeticError, IdaesError):
    """
    IDAES exception to be used when initialization routines fail. All
    initialization routines should raise this Exception if the final step
    fails to converge, and can raise this exception earlier if the routine
    enters a state from which recovery is impossible.
    """


class UserModelError(ValueError, IdaesError):
    """
    IDAES exception for when a user model returns unphysical values that
    prevent further execution of code.
    """



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import sys
import os

sys.path.append(os.path.abspath(".."))  # current folder is ~/examples
from idaes.apps.uncertainty_propagation.uncertainties import (
    quantify_propagate_uncertainty,
)
import pandas as pd
from idaes.apps.uncertainty_propagation.examples.NRTL_model_scripts import (
    NRTL_model,
    NRTL_model_opt,
)

if __name__ == "__main__":
    variable_name = [
        "fs.properties.tau['benzene', 'toluene']",
        "fs.properties.tau['toluene','benzene']",
    ]
    current_path = os.path.dirname(os.path.realpath(__file__))
    data = pd.read_csv(os.path.join(current_path, "BT_NRTL_dataset.csv"))

    def SSE(model, data):
        expr = (
            float(data["vap_benzene"])
            - model.fs.flash.vap_outlet.mole_frac_comp[0, "benzene"]
        ) ** 2 + (
            float(data["liq_benzene"])
            - model.fs.flash.liq_outlet.mole_frac_comp[0, "benzene"]
        ) ** 2
        return expr * 1e4

    results = quantify_propagate_uncertainty(
        NRTL_model, NRTL_model_opt, data, variable_name, SSE
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from copy import deepcopy

import numpy as np

from .unit_cell_lattice import UnitCell, UnitCellLattice
from ..geometry import RectPrism
from ..tiling import CubicTiling
from ..transform_func import ScaleFunc, RotateFunc, ReflectFunc


class PerovskiteLattice(UnitCellLattice):
    RefA = 1
    RefB = 1
    RefC = 1

    # === STANDARD CONSTRUCTOR
    def __init__(self, A, B, C):
        RefUnitCellShape = RectPrism(
            PerovskiteLattice.RefA,
            PerovskiteLattice.RefB,
            PerovskiteLattice.RefC,
            np.array([0, 0, 0], dtype=float),
        )
        RefUnitCellTiling = CubicTiling(RefUnitCellShape)
        RefFracPositions = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.0, 0.5, 0.5]),
        ]
        RefUnitCell = UnitCell(RefUnitCellTiling, RefFracPositions)
        UnitCellLattice.__init__(self, RefUnitCell)
        self._A = PerovskiteLattice.RefA
        self._B = PerovskiteLattice.RefB
        self._C = PerovskiteLattice.RefC
        self.applyTransF(
            ScaleFunc(
                np.array(
                    [
                        A / PerovskiteLattice.RefA,
                        B / PerovskiteLattice.RefB,
                        C / PerovskiteLattice.RefC,
                    ]
                )
            )
        )

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, ScaleFunc):
            self._A *= TransF.Scale[0]
            self._B *= TransF.Scale[1]
            self._C *= TransF.Scale[2]
        UnitCellLattice.applyTransF(self, TransF)

    # === PROPERTY EVALUATION METHODS
    # def isOnLattice(self,P):

    def areNeighbors(self, P1, P2):
        raise NotImplementedError(
            "PerovskiteLattice: there is no universal definition of nearest neighbors."
        )

    def getNeighbors(self, P, layer=1):
        if layer > 2:
            raise ValueError(
                "PerovskiteLattice: there is no universal definition of N-th nearest neighbors."
            )
        RefP = self._getConvertToReference(P)
        if self.isASite(P):
            return []
        elif self.isBSite(P):
            RefNeighs = [
                np.array([0.5, 0.0, 0.0]),
                np.array([0.0, 0.5, 0.0]),
                np.array([0.0, 0.0, 0.5]),
                np.array([-0.5, 0.0, 0.0]),
                np.array([0.0, -0.5, 0.0]),
                np.array([0.0, 0.0, -0.5]),
            ]
        elif self.isOSite(P):
            PointType = self.RefUnitCell.getPointType(self._getConvertToReference(P))
            if PointType == 4:  # i.e., motif aligned with x-axis
                RefNeighs = [
                    np.array([-0.5, 0.0, 0.0]),
                    np.array([0.5, 0.0, 0.0]),
                    np.array([-0.5, 1.0, 0.0]),
                    np.array([-0.5, 0.0, 1.0]),
                    np.array([-0.5, -1.0, 0.0]),
                    np.array([-0.5, 0.0, -1.0]),
                    np.array([0.5, 1.0, 0.0]),
                    np.array([0.5, 0.0, 1.0]),
                    np.array([0.5, -1.0, 0.0]),
                    np.array([0.5, 0.0, -1.0]),
                ]
            elif PointType == 3:  # i.e., motif aligned with y-axis
                RefNeighs = [
                    np.array([0.0, -0.5, 0.0]),
                    np.array([0.0, 0.5, 0.0]),
                    np.array([1.0, -0.5, 0.0]),
                    np.array([0.0, -0.5, 1.0]),
                    np.array([-1.0, -0.5, 0.0]),
                    np.array([0.0, -0.5, -1.0]),
                    np.array([1.0, 0.5, 0.0]),
                    np.array([0.0, 0.5, 1.0]),
                    np.array([-1.0, 0.5, 0.0]),
                    np.array([0.0, 0.5, -1.0]),
                ]
            elif PointType == 2:  # i.e., motif aligned with z-axis
                RefNeighs = [
                    np.array([0.0, 0.0, -0.5]),
                    np.array([0.0, 0.0, 0.5]),
                    np.array([1.0, 0.0, -0.5]),
                    np.array([0.0, 1.0, -0.5]),
                    np.array([-1.0, 0.0, -0.5]),
                    np.array([0.0, -1.0, -0.5]),
                    np.array([1.0, 0.0, 0.5]),
                    np.array([0.0, 1.0, 0.5]),
                    np.array([-1.0, 0.0, 0.5]),
                    np.array([0.0, -1.0, 0.5]),
                ]
            else:
                raise ValueError(
                    "PerovskiteLattice.getNeighbors Should never reach here!"
                )
        else:
            raise ValueError(
                "PerovskiteLattice.getNeighbors given point apparently not on lattice"
            )
        result = deepcopy(RefNeighs)
        for NeighP in result:
            NeighP += RefP
            self._convertFromReference(NeighP)
        return result

    def isASite(self, P):
        return self.RefUnitCell.getPointType(self._getConvertToReference(P)) == 0

    def isBSite(self, P):
        return self.RefUnitCell.getPointType(self._getConvertToReference(P)) == 1

    def isOSite(self, P):
        PointType = self.RefUnitCell.getPointType(self._getConvertToReference(P))
        return PointType == 2 or PointType == 3 or PointType == 4

    def setDesign(self, D, AType, BType, OType):
        for i, P in enumerate(D.Canvas.Points):
            if self.isASite(P):
                D.setContent(i, AType)
            elif self.isBSite(P):
                D.setContent(i, BType)
            elif self.isOSite(P):
                D.setContent(i, OType)
            else:
                raise ValueError("setDesign can not set site not on lattice")

    # === BASIC QUERY METHODS
    @property
    def A(self):
        return self._A

    @property
    def B(self):
        return self._B

    @property
    def C(self):
        return self._C


def getOxygenSymTransFs():
    result = []
    ReflX = ReflectFunc.acrossX()
    ReflY = ReflectFunc.acrossY()
    ReflZ = ReflectFunc.acrossZ()
    RotYY = RotateFunc.fromXYZAngles(0, np.pi, 0)
    RotZZ = RotateFunc.fromXYZAngles(0, 0, np.pi)
    RotX = RotateFunc.fromXYZAngles(np.pi * 0.5, 0, 0)
    RotXX = RotateFunc.fromXYZAngles(np.pi, 0, 0)
    RotXXX = RotateFunc.fromXYZAngles(np.pi * 1.5, 0, 0)

    result.append(RotX)
    result.append(RotXX)
    result.append(RotXXX)

    result.append(ReflX)
    result.append(ReflX + RotX)
    result.append(ReflX + RotXX)
    result.append(ReflX + RotXXX)

    result.append(ReflY)
    result.append(ReflY + RotX)
    result.append(ReflY + RotXX)
    result.append(ReflY + RotXXX)

    result.append(ReflZ)
    result.append(ReflZ + RotX)
    result.append(ReflZ + RotXX)
    result.append(ReflZ + RotXXX)

    result.append(ReflX + ReflY)
    result.append(ReflX + ReflY + RotX)
    result.append(ReflX + ReflY + RotXX)
    result.append(ReflX + ReflY + RotXXX)

    result.append(ReflX + ReflZ)
    result.append(ReflX + ReflZ + RotX)
    result.append(ReflX + ReflZ + RotXX)
    result.append(ReflX + ReflZ + RotXXX)

    result.append(ReflY + ReflZ)
    result.append(ReflY + ReflZ + RotX)
    result.append(ReflY + ReflZ + RotXX)
    result.append(ReflY + ReflZ + RotXXX)

    result.append(ReflX + ReflY + ReflZ)
    result.append(ReflX + ReflY + ReflZ + RotX)
    result.append(ReflX + ReflY + ReflZ + RotXX)
    result.append(ReflX + ReflY + ReflZ + RotXXX)

    result.append(RotYY + RotX)
    result.append(RotYY + RotXX)
    result.append(RotYY + RotXXX)

    result.append(RotYY + ReflX)
    result.append(RotYY + ReflX + RotX)
    result.append(RotYY + ReflX + RotXX)
    result.append(RotYY + ReflX + RotXXX)

    result.append(RotYY + ReflY)
    result.append(RotYY + ReflY + RotX)
    result.append(RotYY + ReflY + RotXX)
    result.append(RotYY + ReflY + RotXXX)

    result.append(RotYY + ReflZ)
    result.append(RotYY + ReflZ + RotX)
    result.append(RotYY + ReflZ + RotXX)
    result.append(RotYY + ReflZ + RotXXX)

    result.append(RotYY + ReflX + ReflY)
    result.append(RotYY + ReflX + ReflY + RotX)
    result.append(RotYY + ReflX + ReflY + RotXX)
    result.append(RotYY + ReflX + ReflY + RotXXX)

    result.append(RotYY + ReflX + ReflZ)
    result.append(RotYY + ReflX + ReflZ + RotX)
    result.append(RotYY + ReflX + ReflZ + RotXX)
    result.append(RotYY + ReflX + ReflZ + RotXXX)

    result.append(RotYY + ReflY + ReflZ)
    result.append(RotYY + ReflY + ReflZ + RotX)
    result.append(RotYY + ReflY + ReflZ + RotXX)
    result.append(RotYY + ReflY + ReflZ + RotXXX)

    result.append(RotYY + ReflX + ReflY + ReflZ)
    result.append(RotYY + ReflX + ReflY + ReflZ + RotX)
    result.append(RotYY + ReflX + ReflY + ReflZ + RotXX)
    result.append(RotYY + ReflX + ReflY + ReflZ + RotXXX)

    result.append(RotZZ + RotX)
    result.append(RotZZ + RotXX)
    result.append(RotZZ + RotXXX)

    result.append(RotZZ + ReflX)
    result.append(RotZZ + ReflX + RotX)
    result.append(RotZZ + ReflX + RotXX)
    result.append(RotZZ + ReflX + RotXXX)

    result.append(RotZZ + ReflY)
    result.append(RotZZ + ReflY + RotX)
    result.append(RotZZ + ReflY + RotXX)
    result.append(RotZZ + ReflY + RotXXX)

    result.append(RotZZ + ReflZ)
    result.append(RotZZ + ReflZ + RotX)
    result.append(RotZZ + ReflZ + RotXX)
    result.append(RotZZ + ReflZ + RotXXX)

    result.append(RotZZ + ReflX + ReflY)
    result.append(RotZZ + ReflX + ReflY + RotX)
    result.append(RotZZ + ReflX + ReflY + RotXX)
    result.append(RotZZ + ReflX + ReflY + RotXXX)

    result.append(RotZZ + ReflX + ReflZ)
    result.append(RotZZ + ReflX + ReflZ + RotX)
    result.append(RotZZ + ReflX + ReflZ + RotXX)
    result.append(RotZZ + ReflX + ReflZ + RotXXX)

    result.append(RotZZ + ReflY + ReflZ)
    result.append(RotZZ + ReflY + ReflZ + RotX)
    result.append(RotZZ + ReflY + ReflZ + RotXX)
    result.append(RotZZ + ReflY + ReflZ + RotXXX)

    result.append(RotZZ + ReflX + ReflY + ReflZ)
    result.append(RotZZ + ReflX + ReflY + ReflZ + RotX)
    result.append(RotZZ + ReflX + ReflY + ReflZ + RotXX)
    result.append(RotZZ + ReflX + ReflY + ReflZ + RotXXX)

    return result



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for solvent condenser unit model.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import (
    check_optimal_termination,
    TransformationFactory,
    ConcreteModel,
    Constraint,
    Param,
    units,
    value,
)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

from idaes.models_extra.column_models.solvent_condenser import SolventCondenser
from idaes.models_extra.column_models.properties.MEA_solvent import (
    configuration as aqueous_mea,
)
from idaes.models_extra.column_models.properties.MEA_vapor import wet_co2
from idaes.core.util.exceptions import InitializationError


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
class TestStripperVaporFlow(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.liquid_properties = GenericParameterBlock(**aqueous_mea)
        m.fs.vapor_properties = GenericParameterBlock(**wet_co2)

        m.fs.unit = SolventCondenser(
            liquid_property_package=m.fs.liquid_properties,
            vapor_property_package=m.fs.vapor_properties,
        )

        m.fs.unit.inlet.flow_mol[0].fix(1.1117)
        m.fs.unit.inlet.temperature[0].fix(339.33)
        m.fs.unit.inlet.pressure[0].fix(184360)
        m.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(0.8817)
        m.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(0.1183)

        m.fs.unit.reflux.flow_mol[0].fix(0.1083)

        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.mass_transfer_term[0, "Vap", "CO2"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.mass_transfer_term[0, "Vap", "H2O"], 100
        )

        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.properties_out[0].pressure, 1e-5
        )
        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.properties_out[0].fug_phase_comp["Vap", "CO2"], 1e-5
        )
        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.properties_out[0].fug_phase_comp["Vap", "H2O"], 1e-4
        )
        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.properties_out[0].temperature, 1e-2
        )
        iscale.set_scaling_factor(
            m.fs.unit.vapor_phase.properties_out[0].enth_mol_phase["Vap"], 1e-3
        )

        iscale.set_scaling_factor(m.fs.unit.vapor_phase.enthalpy_transfer[0], 1e-3)

        iscale.calculate_scaling_factors(m.fs.unit)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):

        assert hasattr(model.fs.unit, "inlet")
        assert len(model.fs.unit.inlet.vars) == 4
        assert hasattr(model.fs.unit.inlet, "flow_mol")
        assert hasattr(model.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.inlet, "temperature")
        assert hasattr(model.fs.unit.inlet, "pressure")

        assert hasattr(model.fs.unit, "reflux")
        assert len(model.fs.unit.reflux.vars) == 4
        assert hasattr(model.fs.unit.reflux, "flow_mol")
        assert hasattr(model.fs.unit.reflux, "mole_frac_comp")
        assert hasattr(model.fs.unit.reflux, "temperature")
        assert hasattr(model.fs.unit.reflux, "pressure")

        assert hasattr(model.fs.unit, "vapor_outlet")
        assert len(model.fs.unit.vapor_outlet.vars) == 4
        assert hasattr(model.fs.unit.vapor_outlet, "flow_mol")
        assert hasattr(model.fs.unit.vapor_outlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.vapor_outlet, "temperature")
        assert hasattr(model.fs.unit.vapor_outlet, "pressure")

        assert isinstance(model.fs.unit.unit_material_balance, Constraint)
        assert isinstance(model.fs.unit.unit_enthalpy_balance, Constraint)
        assert isinstance(model.fs.unit.unit_temperature_equality, Constraint)
        assert isinstance(model.fs.unit.unit_pressure_balance, Constraint)
        assert isinstance(model.fs.unit.zero_flow_param, Param)

        assert number_variables(model.fs.unit) == 55
        assert number_total_constraints(model.fs.unit) == 49
        assert number_unused_variables(model.fs.unit) == 0

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)
        assert_units_equivalent(model.fs.unit.heat_duty[0], units.W)

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.1083, rel=1e-5) == value(
            model.fs.unit.reflux.flow_mol[0]
        )
        assert pytest.approx(0, abs=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0, abs=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(1, rel=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(184360, rel=1e-5) == value(
            model.fs.unit.reflux.pressure[0]
        )
        assert pytest.approx(303.244, rel=1e-5) == value(
            model.fs.unit.reflux.temperature[0]
        )

        assert pytest.approx(1.0034, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.flow_mol[0]
        )
        assert pytest.approx(0.976758, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.0232423, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(184360, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.pressure[0]
        )
        assert pytest.approx(303.244, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.temperature[0]
        )

        assert pytest.approx(-6278.10, rel=1e-5) == value(model.fs.unit.heat_duty[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    - model.fs.unit.reflux.flow_mol[0]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    * model.fs.unit.inlet.mole_frac_comp[0, "CO2"]
                    - model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "CO2"]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                    * model.fs.unit.vapor_outlet.mole_frac_comp[0, "CO2"]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    * model.fs.unit.inlet.mole_frac_comp[0, "H2O"]
                    - model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "H2O"]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                    * model.fs.unit.vapor_outlet.mole_frac_comp[0, "H2O"]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "MEA"]
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    model.fs.unit.vapor_phase.properties_in[0]._enthalpy_flow_term[
                        "Vap"
                    ]
                    - model.fs.unit.vapor_phase.properties_out[0]._enthalpy_flow_term[
                        "Vap"
                    ]
                    - model.fs.unit.liquid_phase[0]._enthalpy_flow_term["Liq"]
                    + model.fs.unit.heat_duty[0]
                )
            )
            <= 1e-6
        )


# -----------------------------------------------------------------------------
class TestStripperHeatDuty(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.liquid_properties = GenericParameterBlock(**aqueous_mea)
        m.fs.vapor_properties = GenericParameterBlock(**wet_co2)

        m.fs.unit = SolventCondenser(
            liquid_property_package=m.fs.liquid_properties,
            vapor_property_package=m.fs.vapor_properties,
        )

        m.fs.unit.inlet.flow_mol[0].fix(1.1117)
        m.fs.unit.inlet.temperature[0].fix(339.33)
        m.fs.unit.inlet.pressure[0].fix(184360)
        m.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(0.8817)
        m.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(0.1183)

        m.fs.unit.heat_duty.fix(-6264)

        xfrm = TransformationFactory("contrib.strip_var_bounds")
        xfrm.apply_to(m, reversible=True)

        # Note: The strip_var_bounds transformation is used to remove the
        # bounds from all the variables in the model. This helps the model
        # converge if the variables are very close to one of their bounds.
        # However, an alternate method would be to specify an appropriate
        # value for the solver option bound_push during model initialization
        # such that the minimum absolute distance from the initial point to
        # the bounds is reduced beyond the default value of 0.01.

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):

        assert hasattr(model.fs.unit, "inlet")
        assert len(model.fs.unit.inlet.vars) == 4
        assert hasattr(model.fs.unit.inlet, "flow_mol")
        assert hasattr(model.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.inlet, "temperature")
        assert hasattr(model.fs.unit.inlet, "pressure")

        assert hasattr(model.fs.unit, "reflux")
        assert len(model.fs.unit.reflux.vars) == 4
        assert hasattr(model.fs.unit.reflux, "flow_mol")
        assert hasattr(model.fs.unit.reflux, "mole_frac_comp")
        assert hasattr(model.fs.unit.reflux, "temperature")
        assert hasattr(model.fs.unit.reflux, "pressure")

        assert hasattr(model.fs.unit, "vapor_outlet")
        assert len(model.fs.unit.vapor_outlet.vars) == 4
        assert hasattr(model.fs.unit.vapor_outlet, "flow_mol")
        assert hasattr(model.fs.unit.vapor_outlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.vapor_outlet, "temperature")
        assert hasattr(model.fs.unit.vapor_outlet, "pressure")

        assert isinstance(model.fs.unit.unit_material_balance, Constraint)
        assert isinstance(model.fs.unit.unit_enthalpy_balance, Constraint)
        assert isinstance(model.fs.unit.unit_temperature_equality, Constraint)
        assert isinstance(model.fs.unit.unit_pressure_balance, Constraint)
        assert isinstance(model.fs.unit.zero_flow_param, Param)

        assert number_variables(model.fs.unit) == 55
        assert number_total_constraints(model.fs.unit) == 49
        assert number_unused_variables(model.fs.unit) == 0

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)
        assert_units_equivalent(model.fs.unit.heat_duty[0], units.W)

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    # @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.108117, rel=1e-5) == value(
            model.fs.unit.reflux.flow_mol[0]
        )
        assert pytest.approx(0, abs=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0, abs=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(1, rel=1e-3) == value(
            model.fs.unit.reflux.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(184360, rel=1e-5) == value(
            model.fs.unit.reflux.pressure[0]
        )
        assert pytest.approx(303.376, rel=1e-5) == value(
            model.fs.unit.reflux.temperature[0]
        )

        assert pytest.approx(1.00358, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.flow_mol[0]
        )
        assert pytest.approx(0.976580, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.0234192, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(184360, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.pressure[0]
        )
        assert pytest.approx(303.376, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.temperature[0]
        )

        assert pytest.approx(-6264, rel=1e-5) == value(model.fs.unit.heat_duty[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    - model.fs.unit.reflux.flow_mol[0]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    * model.fs.unit.inlet.mole_frac_comp[0, "CO2"]
                    - model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "CO2"]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                    * model.fs.unit.vapor_outlet.mole_frac_comp[0, "CO2"]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_mol[0]
                    * model.fs.unit.inlet.mole_frac_comp[0, "H2O"]
                    - model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "H2O"]
                    - model.fs.unit.vapor_outlet.flow_mol[0]
                    * model.fs.unit.vapor_outlet.mole_frac_comp[0, "H2O"]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.reflux.flow_mol[0]
                    * model.fs.unit.reflux.mole_frac_comp[0, "MEA"]
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    model.fs.unit.vapor_phase.properties_in[0]._enthalpy_flow_term[
                        "Vap"
                    ]
                    - model.fs.unit.vapor_phase.properties_out[0]._enthalpy_flow_term[
                        "Vap"
                    ]
                    - model.fs.unit.liquid_phase[0]._enthalpy_flow_term["Liq"]
                    + model.fs.unit.heat_duty[0]
                )
            )
            <= 1e-6
        )

    @pytest.mark.component
    def test_scaling(self, model):
        iscale.set_scaling_factor(
            model.fs.unit.vapor_phase.properties_out[0].fug_phase_comp["Vap", "CO2"],
            1e-5,
        )
        iscale.set_scaling_factor(
            model.fs.unit.vapor_phase.properties_out[0].fug_phase_comp["Vap", "H2O"],
            1e-3,
        )

        iscale.calculate_scaling_factors(model.fs.unit)

        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_material_balance[0, "CO2"]
            )
            == 1
        )
        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_material_balance[0, "H2O"]
            )
            == 1
        )
        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_material_balance[0, "MEA"]
            )
            == 1e8
        )

        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_phase_equilibrium[0, "CO2"]
            )
            == 1e-5
        )
        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_phase_equilibrium[0, "H2O"]
            )
            == 1e-3
        )

        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_temperature_equality[0]
            )
            == 1e-2
        )

        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_enthalpy_balance[0]
            )
            == 1
        )

        assert (
            iscale.get_constraint_transform_applied_scaling_factor(
                model.fs.unit.unit_pressure_balance[0]
            )
            == 1e-5
        )

    @pytest.mark.component
    def test_initialization_error_dof(self, model):
        model.fs.unit.reflux.flow_mol[0].fix(100)

        with pytest.raises(InitializationError):
            model.fs.unit.initialize()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for 0D lumped capacitance heat exchanger model

Author: Rusty Gentile, John Eslick, Andrew Lee
"""
import pytest

from pyomo.environ import (
    ConcreteModel,
    TransformationFactory,
    value,
    units as pyunits,
)

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from idaes.core import FlowsheetBlock

from idaes.models.properties import swco2
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock
from idaes.models.unit_models import (
    HeatExchangerLumpedCapacitance,
    HeatExchangerFlowPattern,
)
from idaes.models.properties.general_helmholtz import helmholtz_available
from idaes.models.unit_models.heat_exchanger import delta_temperature_lmtd_callback
import numpy as np

# Get default solver for testing
solver = get_solver()

# Number of steps for transient simulations
TIME_STEPS = 50


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestHXLCTransientSCO2(object):
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(
            dynamic=True, time_set=[0, 300, 600, 900, 1200, 1500], time_units=pyunits.s
        )

        m.fs.sco2 = swco2.SWCO2ParameterBlock()
        m.fs.fluegas = FlueGasParameterBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            delta_temperature_callback=delta_temperature_lmtd_callback,
            hot_side_name="tube",
            cold_side_name="shell",
            tube={"property_package": m.fs.sco2, "has_pressure_change": True},
            shell={"property_package": m.fs.fluegas, "has_pressure_change": False},
            flow_pattern=HeatExchangerFlowPattern.crossflow,
            dynamic=False,
        )

        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=TIME_STEPS - 1, wrt=m.fs.time, scheme="BACKWARD")

        # Cold-side boundary conditions
        shell_inlet_temperature = 288.15
        shell_flow = 44004.14222
        shell_area = 690073.9153
        shell_hconv = 22
        m.fs.unit.ua_cold_side[:].fix(shell_area * shell_hconv)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "H2O"].fix(0.01027 * shell_flow)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "CO2"].fix(0.000411592 * shell_flow)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "N2"].fix(0.780066026 * shell_flow)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "O2"].fix(0.209252382 * shell_flow)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "NO"].fix(0)
        m.fs.unit.shell_inlet.flow_mol_comp[:, "SO2"].fix(0)
        m.fs.unit.shell_inlet.temperature[:].fix(shell_inlet_temperature)
        m.fs.unit.shell_inlet.pressure[:].fix(101325)

        # Hot-side boundary conditions
        tube_inlet_temperature = 384.35
        tube_inlet_pressure = 7653000
        tube_outlet_pressure = 7500000
        tube_flow = 13896.84163
        tube_area = 19542.2771
        tube_hconv = 1000
        tube_mass = 1160 * 322
        m.fs.unit.ua_hot_side[:].fix(tube_area * tube_hconv)
        tube_inlet_enthalpy = swco2.htpx(
            T=tube_inlet_temperature * pyunits.K, P=tube_inlet_pressure * pyunits.Pa
        )
        m.fs.unit.tube_inlet.flow_mol[:].fix(tube_flow)
        m.fs.unit.tube_inlet.pressure[:].fix(tube_inlet_pressure)
        m.fs.unit.tube_inlet.enth_mol[:].fix(tube_inlet_enthalpy)
        m.fs.unit.tube_outlet.pressure[:].fix(tube_outlet_pressure)

        # Area has no effect on the results here so long as it's not zero
        m.fs.unit.area.fix(1)
        m.fs.unit.crossflow_factor.fix(0.8)
        m.fs.unit.heat_capacity_wall = tube_mass * 466

        m.fs.unit.tube_outlet[:].enth_mol.setub(tube_inlet_enthalpy)
        m.fs.unit.shell_outlet.temperature[:].setlb(shell_inlet_temperature)

        return m

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.integration
    def test_steady_state(self, model):
        """
        Make sure temperatures are all the same for steady state simulation
        """
        model.fs.unit.initialize()
        solver.solve(model)

        # Expected temperatures
        exp_sco2 = np.ones(TIME_STEPS) * 305.2
        exp_air = np.ones(TIME_STEPS) * 370.4
        exp_wall = np.ones(TIME_STEPS) * 339.7

        self.check_temperatures(
            model, np.array(model.fs.time), exp_sco2, exp_air, exp_wall
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.integration
    def test_steps(self, model):
        """
        Add step changes in air temperature similar to the example in the docs
        """
        model.fs.unit.initialize()

        # Add disturbances
        for t in model.fs.time:
            if 300 <= t < 600:
                model.fs.unit.shell_inlet.temperature[t].fix(288.15 - 10)
            elif 600 <= t < 900:
                model.fs.unit.shell_inlet.temperature[t].fix(288.15)
            elif 900 <= t < 1200:
                model.fs.unit.shell_inlet.temperature[t].fix(288.15 + 10)
            elif t >= 1200:
                model.fs.unit.shell_inlet.temperature[t].fix(288.15)

        # Transient solution
        solver.solve(model)

        times = [0, 300, 600, 900, 1200, 1500]
        sco2_exp = [305.2, 304.9, 305.1, 306.5, 305.7, 305.2]
        air_exp = [370.4, 373.1, 370.3, 365.9, 370.7, 370.4]
        wall_exp = [339.4, 338.7, 339.1, 340.7, 339.9, 339.4]

        self.check_temperatures(model, times, sco2_exp, air_exp, wall_exp)

    def check_temperatures(
        self,
        model,
        times,
        expected_sco2_temps,
        expected_air_temps,
        expected_wall_temps,
        tol=1e-3,
    ):
        for i, t in enumerate(times):
            assert value(
                model.fs.unit.tube.properties_out[t].temperature
            ) == pytest.approx(expected_sco2_temps[i], rel=tol)
            assert value(
                model.fs.unit.shell.properties_out[t].temperature
            ) == pytest.approx(expected_air_temps[i], rel=tol)
            assert value(model.fs.unit.temperature_wall[t]) == pytest.approx(
                expected_wall_temps[i], rel=tol
            )



# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General purpose heat integration block for IDAES models
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Plotting and Numpy libraries for composite curves
import matplotlib.pyplot as plt
import numpy as np

# Import Pyomo libraries
from pyomo.environ import Constraint, sqrt, Expression, value, Var

from pyomo.environ import units as pyunits

__author__ = "Alejandro Garciadiego, Alexander Dowling"

# Temperature Epsilon to add or subtract 1 degree to avoid division over
# 0 in equipment not changing temperature
EpsT = 1


def min_utility(blk, heating, cooling, DTmin, eps=1e-6, DG_units=pyunits.Mwatt):
    """
    Function for Duran-Grossmann for minimization of utilities
    This function adds variables and constraints to the flowsheet to
    minimize utilities

    Args:
        blk: flowsheet
        heating: Equipment that requires Heat
        cooling: Equipment from which heat is being removed
        DTmin: HRAT (Heat Recovery Approximation Temperature)
        eps: Epsilon for smoothing operation

    Returns:
        Constraint and variable object representing the calculation
        for hot utility and cold utility
    """

    # Generate lists out of strings from Args
    exchanger_list = heating + cooling
    pinch_streams = exchanger_list

    # Generate dictionaries out of strings to convert to pyomo objects
    pinch_streamsdict = {}
    coolingdict = {}
    heatingdict = {}
    exchangerdict = {}

    for i, el in enumerate(exchanger_list):
        pinch_streamsdict[str(el)] = el

    for i, el in enumerate(cooling):
        coolingdict[str(el)] = el

    for i, el in enumerate(heating):
        heatingdict[str(el)] = el

    for i, el in enumerate(exchanger_list):
        exchangerdict[str(el)] = el

    Q = {}

    for i in exchangerdict:
        Q[i] = pyunits.convert(pinch_streamsdict[i].heat_duty[0], to_units=DG_units)

    # Run function to fill exchanger data for pinch calculations for
    # initialization of variables
    exchangeData = heat_data(blk, heating, cooling, DG_units)

    # Call pinch calculation for initialization of variables in PD class
    PD = pinch_calc(heating, cooling, exchangeData, DTmin, eps)

    # Define dictionary for addition of DTmin to heating equipment
    dT = {}

    for i, el in enumerate(cooling):
        dT[str(el)] = 0

    for i, el in enumerate(heating):
        dT[str(el)] = DTmin

    def T_in(blk, i):
        return pinch_streamsdict[i].control_volume.properties_in[0].temperature

    blk.Tin = Expression(
        pinch_streamsdict.keys(), rule=T_in, doc="Inlet temperature in exchangers"
    )

    def T_out(blk, i):
        return pinch_streamsdict[i].control_volume.properties_out[0].temperature

    blk.Tout = Expression(
        pinch_streamsdict.keys(), rule=T_out, doc="Outlet temperature in exchangers"
    )

    # Expression for cp of equipment with heat exchange
    def Theta_(blk, i):
        if i in heatingdict:
            return Q[i] / (
                0.5
                * (
                    (blk.Tout[i] - blk.Tin[i] + (EpsT))
                    + sqrt((blk.Tout[i] - blk.Tin[i] - (EpsT)) ** 2 + eps)
                )
            )
        else:
            return Q[i] / (
                -0.5
                * (
                    (-(blk.Tout[i] - blk.Tin[i]) - (-EpsT))
                    + sqrt(((-EpsT) - (blk.Tout[i] - blk.Tin[i])) ** 2 + eps)
                )
            )

    blk.Theta = Expression(
        pinch_streamsdict.keys(), rule=Theta_, doc="FCp in exchangers"
    )

    # Define expression for pinch candidate temperature
    def T_(blk, i):
        return pinch_streamsdict[i].control_volume.properties_in[0].temperature + dT[i]

    blk.T_ = Expression(
        pinch_streamsdict.keys(), rule=T_, doc="Pinch candidate temperature"
    )

    # Define variable for heat content above the pinch point
    blk.QAh = Var(
        pinch_streamsdict.keys(),
        initialize=PD.initQAh,
        bounds=(1e-8, None),
        doc="Heat content above pinch",
        units=pyunits.get_units(Q[str(pinch_streams[0])]),
    )

    # Define a constraint to calculate the variable QAh
    def rule_heat_above_pinch(blk, p):
        return blk.QAh[p] == sum(
            blk.Theta[i]
            * (
                0.5
                * ((blk.Tin[i] - blk.T_[p]) + sqrt((blk.Tin[i] - blk.T_[p]) ** 2 + eps))
                - 0.5
                * (
                    (blk.Tout[i] - blk.T_[p])
                    + sqrt((blk.Tout[i] - blk.T_[p]) ** 2 + eps)
                )
            )
            for i in coolingdict
        )

    blk.heat_above_pinch = Constraint(
        pinch_streamsdict.keys(), rule=rule_heat_above_pinch
    )

    # Define variable for heat content below the pinch point
    blk.QAc = Var(
        pinch_streamsdict.keys(),
        initialize=PD.initQAc,
        bounds=(1e-8, None),
        doc="Heat content bellow pinch",
        units=pyunits.get_units(Q[str(pinch_streams[0])]),
    )

    # Define a constraint to calculate the variable QAc
    def rule_heat_below_pinch(blk, p):
        return blk.QAc[p] == sum(
            blk.Theta[i]
            * (
                0.5
                * (
                    (blk.Tout[i] - blk.T_[p] + DTmin)
                    + sqrt((blk.Tout[i] - blk.T_[p] + DTmin) ** 2 + eps)
                )
                - 0.5
                * (
                    (blk.Tin[i] - blk.T_[p] + DTmin)
                    + sqrt((blk.Tin[i] - blk.T_[p] + DTmin) ** 2 + eps)
                )
            )
            for i in heatingdict
        )

    blk.heat_below_pinch = Constraint(
        pinch_streamsdict.keys(), rule=rule_heat_below_pinch
    )

    # Define variable for Heat of hot utility
    blk.Qs = Var(
        initialize=PD.initQs,
        bounds=(1e-8, None),
        doc="Heating utilities",
        units=pyunits.get_units(Q[str(pinch_streams[0])]),
    )

    # Define a constraint to solve for Qs
    # Where 1E-6 is added to both sides of the constraint as a scaling factor
    def rule_heating_utility(blk, p):
        return blk.Qs >= (blk.QAc[p] - blk.QAh[p])

    blk.heating_utility = Constraint(
        pinch_streamsdict.keys(), rule=rule_heating_utility
    )

    # Define variable for Heat of cold utility
    blk.Qw = Var(
        initialize=PD.initQw,
        bounds=(1e-8, None),
        doc="Cooling utilities",
        units=pyunits.get_units(Q[str(pinch_streams[0])]),
    )

    # Define a constraint to solve for Qw
    # Where 1E-6 is added to both sides of the constraint as a scaling factor
    def rule_cooling_utility(blk):
        return blk.Qw == -sum(Q[i] for i in exchangerdict) + blk.Qs

    blk.cooling_utility = Constraint(rule=rule_cooling_utility)


def heat_data(blk, heating, cooling, DG_units=pyunits.Mwatt):
    """
    Function for generating necessary data for Duran-Grossmann initialization
    Allows the calculation of reactors
    Args:
        blk: flowsheet
        heating: Equipment that requires Heat
        cooling: Equipment from which heat is being removed
    Returns:
        Dictionary for the equipment exchanging heat containing:
            Tin
            Tout
            FCp
            Q
    """
    # Generate lists out of strings from Args
    exchanger_list = heating + cooling

    # Generate dictionaries out of strings to convert to pyomo objects
    pinch_streamsdict = {}
    coolingdict = {}
    heatingdict = {}
    exchangerdict = {}

    for i, el in enumerate(exchanger_list):
        pinch_streamsdict[str(el)] = el

    for i, el in enumerate(cooling):
        coolingdict[str(el)] = el

    for i, el in enumerate(heating):
        heatingdict[str(el)] = el

    for i, el in enumerate(exchanger_list):
        exchangerdict[str(el)] = el

    # Generate dictionaries for information for heat data
    Q = {}
    T_in = {}
    T_out = {}
    FCp_ = {}

    # Defining inlet temperature for inlet from equipment control volume
    for i in heatingdict:
        T_in[i] = value(
            pinch_streamsdict[i].control_volume.properties_in[0].temperature
        )
        T_out[i] = (
            value(pinch_streamsdict[i].control_volume.properties_out[0].temperature)
            + EpsT
        )

    # Defining inlet temperature for utlet from equipment control volume
    for i in coolingdict:
        T_in[i] = (
            value(pinch_streamsdict[i].control_volume.properties_in[0].temperature)
            + EpsT
        )
        T_out[i] = value(
            pinch_streamsdict[i].control_volume.properties_out[0].temperature
        )

    # Calculating FCp out of heat in control volume
    # Obtains Q from equipment's control volume heat
    for i, v in pinch_streamsdict.items():
        Q[i] = value(pyunits.convert(v.heat_duty[0], to_units=DG_units))
        FCp_[i] = Q[i] / (T_out[i] - T_in[i])

    # Generate a large dictionary containing all the data obtained
    # from the equipment
    exchangeData = {}
    for i in pinch_streamsdict:
        exchangeData[i] = {
            "T_in": T_in[i],
            "T_out": T_out[i],
            "FCp_": FCp_[i],
            "Q": Q[i],
        }

    return exchangeData


def pinch_calc(heating, cooling, exchangeData, DTmin, eps):
    """
    Function for calculating heat data for Duran-Grossmann initialization

    Args:
        heating: Equipment that requires Heat
        cooling: Equipment from which heat is being removed
        exchangeData: Dictionary containing Tin, Tout, FCp and Q
        for each equipment in lists
        DTmin: HRAT (Heat Recovery Approximation Temperature)
        eps: Epsilon for smoothing operation
    Returns:
        PD (Dictionary containing initialized values for QAh, QAc, Qs and Qw)
    """
    # Generate lists out of strings from Args
    exchanger_list = heating + cooling

    # Generate dictionaries out of strings to convert to pyomo objects
    pinch_streamsdict = {}
    coolingdict = {}
    heatingdict = {}
    exchangerdict = {}

    for i, el in enumerate(exchanger_list):
        pinch_streamsdict[str(el)] = el

    for i, el in enumerate(cooling):
        coolingdict[str(el)] = el

    for i, el in enumerate(heating):
        heatingdict[str(el)] = el

    for i, el in enumerate(exchanger_list):
        exchangerdict[str(el)] = el

    # Generate dictionaries to contain initialized data
    T_ = {}
    initQAh = {}
    initQAc = {}
    b = []

    # Define dictionary for addition of DTmin to heating equipment
    dT = {}
    for i, el in enumerate(cooling):
        dT[str(el)] = 0

    for i, el in enumerate(heating):
        dT[str(el)] = DTmin

    # Calculate pinch temperature candidate
    # Calculate QAh and QAc
    for i in pinch_streamsdict:
        T_[i] = exchangeData[i]["T_in"] + dT[i]
        initQAh[i] = sum(
            exchangeData[j]["FCp_"]
            * (
                0.5
                * (
                    (exchangeData[j]["T_in"] - T_[i])
                    + sqrt((exchangeData[j]["T_in"] - T_[i]) ** 2 + eps)
                )
                - 0.5
                * (
                    (exchangeData[j]["T_out"] - T_[i])
                    + sqrt((exchangeData[j]["T_out"] - T_[i]) ** 2 + eps)
                )
            )
            for j in coolingdict
        )
        initQAc[i] = sum(
            exchangeData[j]["FCp_"]
            * (
                0.5
                * (
                    (exchangeData[j]["T_out"] - T_[i] + DTmin)
                    + sqrt((exchangeData[j]["T_out"] - T_[i] + DTmin) ** 2 + eps)
                )
                - 0.5
                * (
                    (exchangeData[j]["T_in"] - T_[i] + DTmin)
                    + sqrt((exchangeData[j]["T_in"] - T_[i] + DTmin) ** 2 + eps)
                )
            )
            for j in heatingdict
        )

    # Generate array with all possible QS
    for i in exchangerdict:
        b.append(initQAc[i] - initQAh[i])

    # Define largest value of QS
    c = max([max(b), 0.0])
    initQs = c

    # Calculate Qw from largest value of Qs
    initQw = -sum(value(exchangeData[i]["Q"]) for i in exchangerdict) + initQs
    initQw = max([initQw, 0.0])

    # Fill Class with all the data to initialize Duran-Grossmann variables
    PD = PinchDataClass(initQs, initQw)
    PD.initQAh = initQAh
    PD.initQAc = initQAc

    return PD


def generate_curves(CD):
    """
    Function for plotting composite curves

    Args:
        CD: Class containing curve data
    Returns:
        Composite curves
    """
    # Transform information from CD class into arrays for plotting function

    # Transforms cooling equipment data
    CTin = CD.Cooling_Tin
    CTout = CD.Cooling_Tout
    CQ = CD.Cooling_Q
    CTin, CTout, CQ = np.array((CTin, CTout, CQ))

    # Transforms heating equipment data
    HTin = CD.Heating_Tin
    HTout = CD.Heating_Tout
    HQ = CD.Heating_Q
    HTin, HTout, HQ = np.array((HTin, HTout, HQ))

    Thot, Qhot = gen_curves(CTin, CTout, CQ)
    Tcold, Qcold = gen_curves(HTin, HTout, -HQ)
    # Brings Qw scalar into scope out of the class
    Qw = CD.Qw

    plt.figure(figsize=(10, 10))

    # Plot values for Hot streams
    # Negative sign corrects for Q < 0 for hot streams
    plt.plot(-Qhot, Thot, color="r", label="Hot Streams / Streams that are Cooled")

    # Need to shift curve to account for the cooling utility
    Qcold = Qcold + sum(HQ)
    # Plot values for cold streams
    plt.plot(
        (Qcold + value(Qw)),
        Tcold,
        color="b",
        label="Cold Streams / Streams that are Heated",
    )
    plt.xlabel(
        "Cumulative Process-Wide Heat Exchange [" + str(CD.Q_unit) + "]", fontsize=18
    )
    plt.ylabel("Temperature [" + str(CD.T_unit) + "]", fontsize=18)
    plt.title("Composite Curves for minimum utilities", size=20)
    plt.legend(loc="upper left", fontsize=14)
    plt.grid()
    plt.show()
    return


def heat_ex_data(blk, heating, cooling):
    """
    Function for turning IDAES heat exchanging equipment into a class for
    use in plotting

    Args:
        blk: Flowsheet
        heating: Equipment that heats streams
        cooling: Equipment that cools streams
    Returns:
        CD: Class with heating and cooling equipment as arrays
    """
    # Generate dictionaries out of strings to convert to pyomo objects
    exchanger_list = heating + cooling
    pinch_streams = exchanger_list

    pinch_streamsdict = {}
    coolingdict = {}
    heatingdict = {}
    exchangerdict = {}

    T_units = pyunits.get_units(
        pinch_streams[0].control_volume.properties_in[0].temperature
    )
    DG_units = pyunits.get_units(blk.Qw)

    for el in exchanger_list:
        pinch_streamsdict[str(el)] = el
        exchangerdict[str(el)] = el

    for el in cooling:
        coolingdict[str(el)] = el

    for el in heating:
        heatingdict[str(el)] = el

    # Generate zero arrays from length of the cooling list
    CHX = len(coolingdict)
    CTin = np.zeros(CHX)
    CTout = np.zeros(CHX)
    CQ = np.zeros(CHX)

    # Convert Pyomo model values into arrays
    j = 0
    for v in coolingdict.values():
        CTin[j] = value(v.control_volume.properties_in[0].temperature) + 1
        CTout[j] = value(v.control_volume.properties_out[0].temperature)
        CQ[j] = value(pyunits.convert(v.control_volume.heat[0], to_units=DG_units))
        j += 1

    # Generate zero arrays from length of the heating list
    HHX = len(heatingdict)
    HTin = np.zeros(HHX)
    HTout = np.zeros(HHX)
    HQ = np.zeros(HHX)

    # Convert Pyomo model values into arrays
    j = 0
    for v in heatingdict.values():
        HTin[j] = value(v.control_volume.properties_in[0].temperature)
        HTout[j] = value(v.control_volume.properties_out[0].temperature) + 1
        HQ[j] = value(pyunits.convert(v.control_volume.heat[0], to_units=DG_units))
        j += 1

        # Fill class with values and arrays
    Qw = value(blk.Qw)
    CD = CurveData(Qw, T_units, DG_units)
    CD.Cooling_Tin = CTin
    CD.Cooling_Tout = CTout
    CD.Cooling_Q = CQ
    CD.Heating_Tin = HTin
    CD.Heating_Tout = HTout
    CD.Heating_Q = HQ
    return CD


def gen_curves(Tin, Tout, Q):
    """
    Function to do add cumulative heat arrays
    Args:
        Tin: Inlet temperatures of cooling/heating equipment
        Tout: Outlet temperatures of cooling/heating equipment
        Q: Heat of cooling/heating equipment
    Returns:
        Tstar: Cumulative Temperature array
        NQstar: Cumulative heat array
    """
    # Q < 0 for hot streams = heat removing = cooling units
    # Q > 0 for cold streams = heat added = heating units

    # Ignoring edge cases for phase changes
    # Shaping Temperature array
    # Calls unique function to avoid  repeating
    Tstart = unique(np.vstack([Tin, Tout]).reshape((1, -1)))
    Tstar = np.sort(Tstart[0])

    # Generate vector of Tin size and Tunique
    nTin = len(Tin)
    nTstar = len(Tstar)
    Qmat = np.zeros((nTin, nTstar))

    # Generate cumulative arrays for temperature and heat
    for i in range(nTin):
        for j in range(nTstar):
            Qmat[i, j] = linear_interpolation([Tin[i], Tout[i]], [0, Q[i]], Tstar[j])
    Qstar = sum(Qmat, 0)
    NQstar = Qstar.reshape(nTstar)
    return Tstar, NQstar


def linear_interpolation(x, y, t):
    """
    Function to do Linear interpolation with nearest neighbor extrapolation
    Args:
        x: Inlet and Outlet Temperature values
        y: 0 and Heat value
        t: Unique inlet and outlet temperatures
    Returns:
        Qstar: New array of Q Values
    """
    # Set the upper or lower value
    if x[0] < x[1]:
        lo = 0
        hi = 1
    else:
        lo = 1
        hi = 0
    # do Linear interpolation with nearest neighbor extrapolation
    alpha = (x[hi] - t) / (x[hi] - x[lo])
    alpha = max([0, alpha])
    alpha = min([alpha, 1])
    Qsta = alpha * (y[hi] - y[lo]) + y[lo]
    return Qsta


def print_HX_results(blk, exchanger_list):
    """
    Function to print results of Heat Exchangers
    Args:
        blk: flowsheet of the equipment
        exchanger_list: List of equipment to print data
    Returns:
        Printed List
    """
    # Initialize dictionary and fill with exchanger list
    exchangerdict = {}

    for i, el in enumerate(exchanger_list):
        exchangerdict[str(el)] = el

    # initialize null dictionaries for data to be printed
    Tin_ = {}
    Tout_ = {}
    f_ = {}
    Q_ = {}

    # Loop over heat exchangers
    for i, v in exchangerdict.items():
        Tin_[i] = value(v.control_volume.properties_in[0].temperature)
        Tout_[i] = value(v.control_volume.properties_out[0].temperature)
        f_[i] = value(v.control_volume.properties_out[0].flow_mol)
        Q_[i] = value(v.control_volume.heat[0])
        T_units = pyunits.get_units(v.control_volume.properties_in[0].temperature)
        DG_units = pyunits.get_units(v.heat_duty[0])

    # Print the header
    print("Heat Exchanger Summary: ")

    # Print Inlet Temperature, Outlet Temperature and Heat
    for i, v in exchangerdict.items():
        print("Heat exchanger: ", v)
        print(f'Inlet T: {" "*3} {Tin_[i] : 0.3f} {T_units}')
        print(f'Outlet T: {" "*2} {Tout_[i] : 0.3f} {T_units}')
        print(f'Q : {" "*9} {Q_[i]: 0.3f} {DG_units}')


def unique(list1):
    """
    Function to remove not unique elements of a list
    Args:
        list1: List from where to obtain only unique values
    Returns:
        unique_list
    """
    # initialize a null list
    unique_list = []

    # traverse for all elements
    for i in list1:
        # check if exists in unique_list or not
        if i not in unique_list:
            unique_list.append(i)

    return unique_list


class PinchDataClass:
    """
    Class containing all the initialized values for all heat Variables
    """

    def __init__(self, initQs, initQw):
        """
        Args:
            initQs: Initial value of heat of hot utility
            initQw: Initial value of heat to be removed by cold utility
        Returns:
            None
        """
        self.initQs = initQs
        self.initQw = initQw
        self.initQAh = {}
        self.initQAc = {}

    def HeatAbove(self, data_dic):
        """
        Args:
            data_dic: Initial value of heat above pinch
        Returns:
            None
        """
        self.initQAh = data_dic

    def HeatBellow(self, data_dic2):
        """
        Args:
            data_dic2: Initial value of heat below pinch
        Returns:
            None
        """
        self.initQAc = data_dic2


class CurveData:
    """
    Class containing necessary data to generate composite curves
    """

    def __init__(self, Qw, T_unit, Q_unit):
        self.Qw = Qw
        self.T_unit = T_unit
        self.Q_unit = Q_unit

    def Cooling_Tin(self, _list):
        self.Cooling_Tin = _list

    def Cooling_Tout(self, _list):
        self.Cooling_Tout = _list

    def Cooling_Q(self, _list):
        self.Cooling_Q = _list

    def Heating_Tin(self, _list):
        self.Heating_Tin = _list

    def Heating_Tout(self, _list):
        self.Heating_Tout = _list

    def Heating_Q(self, _list):
        self.Heating_Q = _list



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from .base.generic_property import GenericParameterBlock, GenericStateBlock
from .base.generic_reaction import GenericReactionParameterBlock, GenericReactionBlock



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests get environment info
"""
import json
import pytest
import idaes.ver as ver
from idaes.core.util.env_info import EnvironmentInfo


@pytest.mark.unit
def test_env_info():
    x = EnvironmentInfo()
    d = x.to_dict()
    assert "IDAES" in d
    assert "Pyomo" in d
    assert "OS" in d
    assert hasattr(x, "package_version")
    assert x.version_string == ver.__version__
    s = x.to_json()
    d = json.loads(s)
    assert "IDAES" in d
    assert "Pyomo" in d
    assert "OS" in d



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

import time

# Need to time imports, so must start timer before importing
# pylint: disable=wrong-import-position
_command_import_start_time = time.time()

import pkgutil
import click
from idaes.commands.base import command_base as cb

# import all the commands
for finder, dotted_module_name, is_pkg in pkgutil.walk_packages(__path__):
    module_name_parts = dotted_module_name.split(".")
    module_name = module_name_parts[-1]
    is_test_module = module_name.startswith("test_")
    if dotted_module_name == "base":
        pass
    elif is_test_module:
        pass
    else:
        try:
            finder.find_spec(module_name).loader.load_module(module_name)
        except ModuleNotFoundError:
            click.echo(
                f"Could not import commands from {module_name}. Perhaps a dependency is missing."
            )

_command_import_total_time = time.time() - _command_import_start_time



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
from pyomo.common import unittest as pyo_unittest
from idaes.apps.grid_integration.model_data import (
    ThermalGeneratorModelData,
    RenewableGeneratorModelData,
)


@pytest.fixture
def generator_params():
    return {
        "gen_name": "Testing_Generator",
        "bus": "bus5",
        "p_min": 30,
        "p_max": 76,
        "min_down_time": 2,
        "min_up_time": 3,
        "ramp_up_60min": 100,
        "ramp_down_60min": 100,
        "shutdown_capacity": 30,
        "startup_capacity": 30,
        "initial_status": 4,
        "initial_p_output": 30,
        "production_cost_bid_pairs": [(30, 25), (45, 23), (60, 27), (76, 35)],
        "startup_cost_pairs": [(2, 1000), (6, 1500), (10, 2000)],
        "fixed_commitment": None,
    }


@pytest.fixture
def renewable_generator_params():
    return {
        "gen_name": "Testing_Renewable_Generator",
        "bus": "bus5",
        "p_min": 0,
        "p_max": 200,
        "p_cost": 0,
        "fixed_commitment": None,
    }


@pytest.fixture
def generator_data_object(generator_params):
    return ThermalGeneratorModelData(**generator_params)


@pytest.fixture
def renewable_generator_data_object(renewable_generator_params):
    return RenewableGeneratorModelData(**renewable_generator_params)


@pytest.mark.unit
def test_create_thermal_model_data_object(generator_params, generator_data_object):

    # test scalar values
    for param_name in generator_params:
        if param_name not in ["production_cost_bid_pairs", "startup_cost_pairs"]:
            param_val = getattr(generator_data_object, param_name)
            assert param_val == generator_params[param_name]

    # test bids
    attr_names = ["p_cost", "startup_cost"]
    param_names = ["production_cost_bid_pairs", "startup_cost_pairs"]

    for a_name, p_name in zip(attr_names, param_names):
        bids = getattr(generator_data_object, a_name)
        expected_bid = [(key, val) for key, val in generator_params[p_name]]
        pyo_unittest.assertStructuredAlmostEqual(first=bids, second=expected_bid)


@pytest.mark.unit
def test_create_renewable_model_data_object(
    renewable_generator_params, renewable_generator_data_object
):

    for name, value in renewable_generator_data_object:
        if name == "generator_type":
            assert value == "renewable"
        else:
            assert renewable_generator_params[name] == value


@pytest.mark.unit
@pytest.mark.parametrize(
    "param_name, value",
    [
        ("p_min", "30"),
        ("p_max", "76"),
        ("min_down_time", "10"),
        ("min_up_time", "100"),
        ("ramp_up_60min", "100"),
        ("ramp_down_60min", "-100"),
        ("shutdown_capacity", "100"),
        ("startup_capacity", "100"),
    ],
)
def test_create_model_data_with_non_real_numbers(param_name, value, generator_params):
    generator_params[param_name] = value
    with pytest.raises(
        TypeError, match=f"Value for {param_name} shoulde be real numbers."
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
@pytest.mark.parametrize(
    "param_name, value",
    [
        ("p_min", -30),
        ("min_down_time", -10),
        ("min_up_time", -10),
        ("ramp_up_60min", -0.01),
        ("ramp_down_60min", -100),
    ],
)
def test_create_model_data_with_negative_values(param_name, value, generator_params):
    generator_params[param_name] = value
    with pytest.raises(ValueError, match="Value should be greater than or equal to 0."):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
@pytest.mark.parametrize(
    "param_name, value",
    [("p_max", 25), ("shutdown_capacity", 10), ("startup_capacity", 15)],
)
def test_create_model_data_with_less_than_pmin_data(
    param_name, value, generator_params
):
    generator_params[param_name] = value
    with pytest.raises(
        ValueError, match=f"Value for {param_name} shoulde be greater or equal to Pmin."
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_invalid_fixed_commitment_value(generator_params):
    generator_params["fixed_commitment"] = 5
    with pytest.raises(
        ValueError, match=r"^(Value for generator fixed commitment must be one of)"
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_invalid_generator_name(generator_params):
    generator_params["gen_name"] = 102_111
    with pytest.raises(TypeError, match=r".*gen_name must be str.*"):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_invalid_bus_name(generator_params):
    generator_params["bus"] = 102
    with pytest.raises(TypeError, match=r".*bus must be str.*"):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
@pytest.mark.parametrize(
    "param_name", ["production_cost_bid_pairs", "startup_cost_pairs"]
)
def test_empty_bid_pairs(param_name, generator_params):
    generator_params[param_name] = None
    with pytest.raises(
        ValueError, match=r"Empty production|startup cost pairs are provided."
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_bid_missing_pmin(generator_params):
    generator_params["production_cost_bid_pairs"] = generator_params[
        "production_cost_bid_pairs"
    ][1:]
    with pytest.raises(
        ValueError, match=r"^(The first power output in the bid should be the Pmin)"
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_bid_missing_pmax(generator_params):
    generator_params["production_cost_bid_pairs"] = [(30, 0), (45, 0)]
    with pytest.raises(
        ValueError, match=r"^(The last power output in the bid should be the Pmax)"
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_invalid_start_up_bid(generator_params):
    generator_params["startup_cost_pairs"] = [(0, 0)]
    with pytest.raises(
        ValueError,
        match=r"^(The first startup lag should be the same as minimum down time)",
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_invalid_pcost(generator_params):
    generator_params["production_cost_bid_pairs"] = [(30, 0)]
    with pytest.raises(
        ValueError,
        match=r"^(A valid production_cost_bid_pairs requires at least 2 points)",
    ):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
def test_model_data_iterator(generator_data_object):

    expected_param_names = [
        "gen_name",
        "bus",
        "p_min",
        "p_max",
        "min_down_time",
        "min_up_time",
        "ramp_up_60min",
        "ramp_down_60min",
        "shutdown_capacity",
        "startup_capacity",
        "p_cost",
        "startup_cost",
        "fixed_commitment",
        "generator_type",
        "initial_status",
        "initial_p_output",
        "include_default_p_cost",
    ]
    iter_result = [name for name, value in generator_data_object]

    expected_param_names.sort()
    iter_result.sort()

    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_param_names, second=iter_result
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "value, error, msg",
    [
        ("1", TypeError, "Value for initial_status shoulde be real numbers"),
        (0, ValueError, "Value for initial_status cannot be zero"),
    ],
)
def test_invalid_initial_status(value, error, msg, generator_params):

    generator_params["initial_status"] = value
    with pytest.raises(error, match=msg):
        ThermalGeneratorModelData(**generator_params)


@pytest.mark.unit
@pytest.mark.parametrize(
    "initial_status, initial_p_output, error, msg",
    [
        (9, "1", TypeError, "Value for initial_p_output shoulde be real numbers"),
        (
            9,
            29,
            ValueError,
            r".*so the initial power output should at least be p_min.*",
        ),
        (-9, 30, ValueError, r".*so the initial power output should at 0.*"),
    ],
)
def test_invalid_initial_p_output(
    initial_status, initial_p_output, error, msg, generator_params
):

    generator_params["initial_p_output"] = initial_p_output
    generator_params["initial_status"] = initial_status
    with pytest.raises(error, match=msg):
        ThermalGeneratorModelData(**generator_params)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np

from ..geometry import RectPrism
from ..atom import Atom


def readPointsAndAtomsFromCFG(filename):
    Pts = []
    Atoms = []
    GS = 1.0
    Vx = np.array([0, 0, 0], dtype=float)
    Vy = np.array([0, 0, 0], dtype=float)
    Vz = np.array([0, 0, 0], dtype=float)
    blnAtomToAdd = False  # Used to track
    Elem = None  # Used to store most recent element in the new format
    with open(filename, "r") as infile:
        for line in infile:
            splitLine = line.split()
            if len(splitLine) == 0:
                continue
            blnAtomToAdd = False
            s1 = s2 = s3 = None
            Tagname = splitLine[0]
            if Tagname[0] == "#":
                continue
            elif Tagname == "A":
                GS = GS * float(splitLine[2])
            elif Tagname == "H0(1,1)":
                Vx[0] = GS * float(splitLine[2])
            elif Tagname == "H0(1,2)":
                Vx[1] = GS * float(splitLine[2])
            elif Tagname == "H0(1,3)":
                Vx[2] = GS * float(splitLine[2])
            elif Tagname == "H0(2,1)":
                Vy[0] = GS * float(splitLine[2])
            elif Tagname == "H0(2,2)":
                Vy[1] = GS * float(splitLine[2])
            elif Tagname == "H0(2,3)":
                Vy[2] = GS * float(splitLine[2])
            elif Tagname == "H0(3,1)":
                Vz[0] = GS * float(splitLine[2])
            elif Tagname == "H0(3,2)":
                Vz[1] = GS * float(splitLine[2])
            elif Tagname == "H0(3,3)":
                Vz[2] = GS * float(splitLine[2])
            elif Tagname[0].isdigit() and len(splitLine) == 1:
                # Start of new format, line for mol weight
                line = next(infile)
                splitLine = line.split()
                Elem = Atom(splitLine[0])
            elif Tagname[0].isdigit() and splitLine[1][0].isdigit():
                # New format atom
                blnAtomToAdd = True
                s1 = float(splitLine[0])
                s2 = float(splitLine[1])
                s3 = float(splitLine[2])
            elif Tagname[0].isdigit() and splitLine[1][0].isalpha():
                # Old format atom
                blnAtomToAdd = True
                Elem = Atom(splitLine[1])
                s1 = float(splitLine[2])
                s2 = float(splitLine[3])
                s3 = float(splitLine[4])
            else:
                # Other entry, not useful
                pass
            if blnAtomToAdd:
                assert Elem is not None
                x = s1 * Vx[0] + s2 * Vy[0] + s3 * Vz[0]
                y = s1 * Vx[1] + s2 * Vy[1] + s3 * Vz[1]
                z = s1 * Vx[2] + s2 * Vy[2] + s3 * Vz[2]
                Pts.append(np.array([x, y, z], dtype=float))
                Atoms.append(Elem)
    return Pts, Atoms


def writeDesignToCFG(
    D, filename, GS=None, BBox=None, AuxPropMap=None, blnGroupByType=True
):
    if BBox is None:
        BBox = RectPrism.fromPointsBBox(D.Canvas.Points)
        BBox.scale(2.0)
    with open(filename, "w") as outfile:
        outfile.write("Number of particles = {}\n".format(D.NonVoidCount))
        outfile.write(
            "A = {} Angstrom (basic length-scale)\n".format(1.0 if GS is None else GS)
        )
        outfile.write("H0(1,1) = {} A\n".format(BBox.Vx[0]))
        outfile.write("H0(1,2) = {} A\n".format(BBox.Vx[1]))
        outfile.write("H0(1,3) = {} A\n".format(BBox.Vx[2]))
        outfile.write("H0(2,1) = {} A\n".format(BBox.Vy[0]))
        outfile.write("H0(2,2) = {} A\n".format(BBox.Vy[1]))
        outfile.write("H0(2,3) = {} A\n".format(BBox.Vy[2]))
        outfile.write("H0(3,1) = {} A\n".format(BBox.Vz[0]))
        outfile.write("H0(3,2) = {} A\n".format(BBox.Vz[1]))
        outfile.write("H0(3,3) = {} A\n".format(BBox.Vz[2]))
        outfile.write(".NO_VELOCITY.\n")
        if AuxPropMap is not None:
            outfile.write("entry_count = {}\n".format(len(AuxPropMap) + 3))
            for i, AuxProp in enumerate(AuxPropMap):
                PropName = AuxProp[0]
                PropUnit = AuxProp[1]
                outfile.write("auxiliary[{}] = {} [{}]".format(i, PropName, PropUnit))
            outfile.write("\n")
        else:
            outfile.write("entry_count = 3\n")

        if blnGroupByType:
            for Elem in D.NonVoidElems:
                outfile.write("{}\n".format(Elem.Mass))
                outfile.write("{}\n".format(Elem.Symbol))
                for i in range(len(D)):
                    if D.Contents[i] == Elem:
                        P = BBox.getFractionalCoords(D.Canvas.Points[i])
                        outfile.write("{} {} {}".format(P[0], P[1], P[2]))
                        if AuxPropMap is not None:
                            for AuxProp in AuxPropMap:
                                if i in AuxPropMap[AuxProp]:
                                    outfile.write(" {}".format(AuxPropMap[AuxProp][i]))
                        outfile.write("\n")
        else:
            for i in range(len(D)):
                if not (D.Contents[i] is None or D.Contents[i] == Atom()):
                    P = BBox.getFractionalCoords(D.Canvas.Points[i])
                    Elem = D.Contents[i]
                    outfile.write(
                        "{} {} {} {} {}".format(
                            Elem.Mass, Elem.Symbol, P[0], P[1], P[2]
                        )
                    )
                    if AuxPropMap is not None:
                        for AuxProp in AuxPropMap:
                            if i in AuxPropMap[AuxProp]:
                                outfile.write(" {}".format(AuxPropMap[AuxProp][i]))
                    outfile.write("\n")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Two-dimensional model of a slab of some conductive material. Presently it is
used to for the SOC electrolyte, but it can also be used for the interconnect.
It is discretized using a finite volume method with centered differences for
interpolation, and contains only thermal transport and resistive heating terms.

Boundary variables:
    - ``temperature_deviation_x0[t, iz]``
    - ``temperature_deviation_x1[t, iz]``
    - ``heat_flux_x0[t, iz]``
    - ``heat_flux_x1[t, iz]``

Instances of ``Var`` that must be fixed:
    - ``length_x``: Thickness of conductive slab
    - ``heat_capacity``: Specific heat capacity of slab on a mass-basis (i.e., in J/(kg * K))
    - ``density``: Mass density of slab
    - ``thermal_conductivity``: Thermal conductivity of slab
    - ``resistivity_log_preexponential_factor``: Natural logarithm of resistivity preexponential factor in ohm * m
    - ``resistivity_thermal_exponent_dividend``: Parameter divided by temperature in resistivity equation, in K.
      Would be something like (reduced) activation energy, but it can be both negative and positive.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "John Eslick, Douglas Allan"

from pyomo.common.config import ConfigValue, Bool
from pyomo.dae import DerivativeVar
import pyomo.environ as pyo


from idaes.core import declare_process_block_class, UnitModelBlockData
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog


@declare_process_block_class("SocConductiveSlab")
class SocConductiveSlabData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "control_volume_xfaces",
        ConfigValue(
            description="List containing coordinates of control volume faces "
            "in x direction. Coordinates must start with zero, be strictly "
            "increasing, and end with one"
        ),
    )
    CONFIG.declare(
        "voltage_drop_custom",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, add voltage_drop_custom Var to be connected to degradation models",
        ),
    )

    common._submodel_boilerplate_config(CONFIG)
    common._thermal_boundary_conditions_config(CONFIG, thin=False)

    def build(self):
        super().build()

        # Set up some sets for the space and time indexing
        dynamic = self.config.dynamic
        tset = self.flowsheet().config.time
        # Set up node and face sets and get integer indices for them
        izfaces, iznodes = common._face_initializer(
            self, self.config.control_volume_zfaces, "z"
        )
        ixfaces, ixnodes = common._face_initializer(
            self, self.config.control_volume_xfaces, "x"
        )
        # Channel thickness AKA length in the x direction is specific to the
        # channel so local variable here is the only option
        self.length_x = pyo.Var(
            doc="Thickness of the electrode (x-direction)",
            units=pyo.units.m,
        )

        common._submodel_boilerplate_create_if_none(self)
        common._create_thermal_boundary_conditions_if_none(self, thin=False)

        self.temperature_deviation_x = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            initialize=0,
            doc="Temperature at node centers",
            units=pyo.units.K,
        )
        if self.config.has_holdup:
            self.int_energy_density_solid = pyo.Var(
                tset,
                ixnodes,
                iznodes,
                doc="Internal energy density of solid electrode",
                units=pyo.units.J / pyo.units.m**3,
            )
        self.resistivity_log_preexponential_factor = pyo.Var(
            doc="Logarithm of resistivity preexponential factor " "in units of ohm*m",
            units=pyo.units.dimensionless,
        )
        self.resistivity_thermal_exponent_dividend = pyo.Var(
            doc="Parameter divided by temperature in exponential", units=pyo.units.K
        )
        if self.config.voltage_drop_custom:
            self.voltage_drop_custom = pyo.Var(
                tset,
                ixnodes,
                iznodes,
                units=pyo.units.volts,
                doc="Custom voltage drop term for degradation modeling",
            )

        # Parameters
        self.heat_capacity = pyo.Var(
            initialize=200,
            doc="Heat capacity of slab (mass basis)",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        self.density = pyo.Var(
            initialize=1000,
            doc="Density of slab",
            units=pyo.units.kg / pyo.units.m**3,
        )
        self.thermal_conductivity = pyo.Var(
            initialize=80,
            doc="Thermal conductivity of slab",
            units=pyo.units.W / pyo.units.m / pyo.units.K,
        )

        @self.Expression(tset, ixnodes, iznodes)
        def temperature(b, t, ix, iz):
            if b.config.include_temperature_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x[t, ix, iz]
            else:
                return b.temperature_z[t, iz]

        if self.config.has_holdup:

            @self.Constraint(tset, ixnodes, iznodes)
            def int_energy_density_solid_eqn(b, t, ix, iz):
                return b.int_energy_density_solid[
                    t, ix, iz
                ] == b.heat_capacity * b.density * (
                    b.temperature[t, ix, iz] - 1000 * pyo.units.K
                )

        if dynamic:
            self.dcedt_solid = DerivativeVar(
                self.int_energy_density_solid,
                wrt=tset,
                initialize=0,
                doc="Internal energy density time derivative",
            )
        else:
            self.dcedt_solid = pyo.Param(
                tset,
                ixnodes,
                iznodes,
                initialize=0,
                units=pyo.units.W / pyo.units.m**3,
            )

        @self.Expression(iznodes)
        def dz(b, iz):
            return b.zfaces.at(iz + 1) - b.zfaces.at(iz)

        @self.Expression(ixnodes)
        def dx(b, ix):
            return b.xfaces.at(ix + 1) - b.xfaces.at(ix)

        @self.Expression(ixnodes, iznodes)
        def node_volume(b, ix, iz):
            return (
                b.length_x[None]
                * b.length_y[None]
                * b.length_z[None]
                * b.dz[iz]
                * b.dx[ix]
            )

        @self.Expression(ixnodes)
        def zface_area(b, ix):
            return b.length_y[None] * b.length_x[None] * b.dx[ix]

        @self.Expression(iznodes)
        def xface_area(b, iz):
            return b.length_y[None] * b.length_z[None] * b.dz[iz]

        @self.Expression(tset, iznodes)
        def current(b, t, iz):
            return b.current_density[t, iz] * b.xface_area[iz]

        @self.Expression(tset, ixfaces, iznodes)
        def dTdx(b, t, ix, iz):
            return common._interpolate_2D(
                ic=ix,
                ifaces=ixfaces,
                nodes=b.xnodes,
                faces=b.xfaces,
                phi_func=lambda ixf: b.temperature_deviation_x[t, ixf, iz] / b.length_x,
                phi_bound_0=(
                    b.temperature_deviation_x0[t, iz]
                    - b.temperature_deviation_x[t, ixnodes.first(), iz]
                )
                / (b.xfaces.first() - b.xnodes.first())
                / b.length_x,
                phi_bound_1=(
                    b.temperature_deviation_x[t, ixnodes.last(), iz]
                    - b.temperature_deviation_x1[t, iz]
                )
                / (b.xnodes.last() - b.xfaces.last())
                / b.length_x,
                derivative=True,
            )

        @self.Expression(tset, ixnodes, izfaces)
        def dTdz(b, t, ix, iz):
            return common._interpolate_2D(
                ic=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda izf: b.temperature[t, ix, izf] / b.length_z[None],
                phi_bound_0=0 * pyo.units.K / pyo.units.m,
                phi_bound_1=0 * pyo.units.K / pyo.units.m,
                derivative=True,
            )

        @self.Expression(tset, ixfaces, iznodes)
        def heat_flux_x(b, t, ix, iz):
            return -b.thermal_conductivity * b.dTdx[t, ix, iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x0_eqn(b, t, iz):
            return b.heat_flux_x0[t, iz] == b.heat_flux_x[t, ixfaces.first(), iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x1_eqn(b, t, iz):
            return b.heat_flux_x1[t, iz] == b.heat_flux_x[t, ixfaces.last(), iz]

        @self.Expression(tset, ixnodes, izfaces)
        def heat_flux_z(b, t, ix, iz):
            return -b.thermal_conductivity * b.dTdz[t, ix, iz]

        @self.Expression(tset, ixnodes, iznodes)
        def resistivity(b, t, ix, iz):
            return (
                pyo.units.ohm
                * pyo.units.m
                * pyo.exp(
                    b.resistivity_log_preexponential_factor
                    + b.resistivity_thermal_exponent_dividend / b.temperature[t, ix, iz]
                )
            )

        @self.Expression(tset, ixnodes, iznodes)
        def resistance(b, t, ix, iz):
            return b.resistivity[t, ix, iz] * b.length_x * b.dx[ix] / b.xface_area[iz]

        @self.Expression(tset, ixnodes, iznodes)
        def voltage_drop(b, t, ix, iz):
            if self.config.voltage_drop_custom:
                return (
                    b.current[t, iz] * b.resistance[t, ix, iz]
                    + b.voltage_drop_custom[t, ix, iz]
                )
            else:
                return b.current[t, iz] * b.resistance[t, ix, iz]

        @self.Expression(tset, iznodes)
        def resistance_total(b, t, iz):
            return sum(b.resistance[t, ix, iz] for ix in ixnodes)

        @self.Expression(tset, iznodes)
        def voltage_drop_total(b, t, iz):
            return sum(b.voltage_drop[t, ix, iz] for ix in ixnodes)

        @self.Expression(tset, ixnodes, iznodes)
        def joule_heating(b, t, ix, iz):
            # current_density is the current density so have to multiply it be Area I**2 = i**2*A**2
            # R = rho * dx / Area / (1-porosity) heating = I**2*R
            return b.current[t, iz] * b.voltage_drop[t, ix, iz]

        @self.Constraint(tset, ixnodes, iznodes)
        def energy_balance_solid_eqn(b, t, ix, iz):
            return (
                b.node_volume[ix, iz] * b.dcedt_solid[t, ix, iz]
                == b.xface_area[iz]
                * (b.heat_flux_x[t, ix, iz] - b.heat_flux_x[t, ix + 1, iz])
                + b.zface_area[ix]
                * (b.heat_flux_z[t, ix, iz] - b.heat_flux_z[t, ix, iz + 1])
                + b.joule_heating[t, ix, iz]
            )

        if dynamic:
            self.energy_balance_solid_eqn[tset.first(), :, :].deactivate()

    def calculate_scaling_factors(self):
        pass

    def initialize_build(self, outlvl=idaeslog.NOTSET, solver=None, optarg=None):
        # We've got initial estimates of temperature and heat flux at both ends of the slab. The fluxes are actually the
        # important things to preserve, but we cannot fix both of them (BVPs with two sets of Neumann boundary
        # conditions are ill-defined).
        raise NotImplementedError(
            "Initialization for the conductive_slab unit model is not implemented because there is no obvious set "
            "of boundary conditions to fix during solid_oxide_cell initialization, and it is not meant to be "
            "initialized in isolation."
        )

    def model_check(self):
        pass

    def recursive_scaling(self):
        gsf = iscale.get_scaling_factor

        def ssf(c, s):
            iscale.set_scaling_factor(c, s, overwrite=False)

        sgsf = iscale.set_and_get_scaling_factor

        def cst(c, s):
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        sk = 1  # Thermal conductivity is ~1
        sLx = sgsf(self.length_x, len(self.ixnodes) / self.length_x.value)
        # sLy = sgsf(self.length_y,1/self.length_y[None].value)
        # sLz = sgsf(self.length_z,len(self.iznodes)/self.length_z[None].value)
        sLy = 1 / self.length_y[None].value
        sLz = len(self.iznodes) / self.length_z[None].value

        for t in self.flowsheet().time:
            for iz in self.iznodes:
                if self.heat_flux_x0[t, iz].is_reference():
                    sq0 = gsf(self.heat_flux_x0[t, iz].referent, default=1e-2)
                else:
                    sq0 = sgsf(self.heat_flux_x0[t, iz], 1e-2)
                cst(self.heat_flux_x0_eqn[t, iz], sq0)
                if not self.temperature_deviation_x0.is_reference():
                    ssf(self.temperature_deviation_x0, sq0 * sLx / sk)

                if self.heat_flux_x1[t, iz].is_reference():
                    sq1 = gsf(self.heat_flux_x1[t, iz].referent, default=1e-2)
                else:
                    sq1 = sgsf(self.heat_flux_x1[t, iz], 1e-2)
                cst(self.heat_flux_x1_eqn[t, iz], sq1)
                if not self.temperature_deviation_x1.is_reference():
                    ssf(self.temperature_deviation_x1, sq1 * sLx / sk)

                sqx = min(sq0, sq1)
                # sqz = 10 * sqx  # Heuristic

                for ix in self.ixnodes:
                    sDT = sgsf(self.temperature_deviation_x[t, ix, iz], sqx * sLx / sk)

                    if self.config.has_holdup:
                        s_rho_U_solid = sgsf(
                            self.int_energy_density_solid[t, ix, iz],
                            1 / (self.heat_capacity.value * self.density.value * sDT),
                        )
                        cst(self.int_energy_density_solid_eqn[t, ix, iz], s_rho_U_solid)

                    cst(self.energy_balance_solid_eqn[t, ix, iz], sqx * sLy * sLz)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions for use in testing IDAES models.
"""
# These are methods for testing, and break a few conventions
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access

__author__ = "Andrew Lee"


import os
from typing import Callable, Union

from pyomo.environ import Constraint, Set, units, Var
from pyomo.common.config import ConfigBlock
from pyomo.common import Executable
from pyomo.common.dependencies import attempt_import

from idaes.core import (
    declare_process_block_class,
    PhysicalParameterBlock,
    StateBlock,
    StateBlockData,
    ReactionParameterBlock,
    ReactionBlockBase,
    ReactionBlockDataBase,
    MaterialFlowBasis,
    MaterialBalanceType,
    EnergyBalanceType,
    Component,
    Phase,
)

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def initialization_tester(m, dof=0, unit=None, **init_kwargs):
    """
    A method to test initialization methods on IDAES models. This method is
    designed to be used as part of the tests for most models.

    This method checks that the initialization methods runs as expected
    and that the state of the model (active/inactive and fixed/unfixed)
    remains the same.

    This method also add some dummy constraints to the model and deactivates
    them to make sure that the initialization does not affect their status.

    Args:
        m: a Concrete model which contains a flowsheet and a model named unit
            (i.e. m.fs.unit) which will be initialized
        dof: expected degrees of freedom during initialization, default=0
        unit: unit object to test, if None assume m.fs.unit, default='None'
        init_kwargs: model specific arguments to pass to initialize method
                     (e.g. initial guesses for states)

    Returns:
        None

    Raises:
        AssertionErrors if an issue is found
    """
    if unit is None:
        unit = m.fs.unit
    # Add some extra constraints and deactivate them to make sure
    # they remain deactivated
    # Test both indexed and unindexed constraints
    unit.__dummy_var = Var()
    unit.__dummy_equality = Constraint(expr=unit.__dummy_var == 5)
    unit.__dummy_inequality = Constraint(expr=unit.__dummy_var <= 10)
    unit.__dummy_set = Set(initialize=[1])

    def deq_idx(b, i):
        return unit.__dummy_var == 5

    unit.__dummy_equality_idx = Constraint(unit.__dummy_set, rule=deq_idx)

    def dieq_idx(b, i):
        return unit.__dummy_var <= 10

    unit.__dummy_inequality_idx = Constraint(unit.__dummy_set, rule=dieq_idx)

    unit.__dummy_equality.deactivate()
    unit.__dummy_inequality.deactivate()
    unit.__dummy_equality_idx[1].deactivate()
    unit.__dummy_inequality_idx[1].deactivate()

    orig_fixed_vars = fixed_variables_set(m)
    orig_act_consts = activated_constraints_set(m)

    unit.initialize(**init_kwargs)

    assert degrees_of_freedom(m) == dof

    fin_fixed_vars = fixed_variables_set(m)
    fin_act_consts = activated_constraints_set(m)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars

    # Check dummy constraints and clean up
    assert not unit.__dummy_equality.active
    assert not unit.__dummy_inequality.active
    assert not unit.__dummy_equality_idx[1].active
    assert not unit.__dummy_inequality_idx[1].active

    unit.del_component(unit.__dummy_inequality)
    unit.del_component(unit.__dummy_equality)
    unit.del_component(unit.__dummy_inequality_idx)
    unit.del_component(unit.__dummy_equality_idx)
    unit.del_component(unit.__dummy_var)
    unit.del_component(unit.__dummy_set)


# -----------------------------------------------------------------------------
# Define some generic PhysicalBlock and ReactionBlock classes for testing
@declare_process_block_class("PhysicalParameterTestBlock")
class _PhysicalParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PhysicalParameterBlock, self).build()

        self.p1 = Phase()
        self.p2 = Phase()

        self.c1 = Component()
        self.c2 = Component()

        self.phase_equilibrium_idx = Set(initialize=["e1", "e2"])
        self.element_list = Set(initialize=["H", "He", "Li"])
        self.element_comp = {
            "c1": {"H": 1, "He": 2, "Li": 3},
            "c2": {"H": 4, "He": 5, "Li": 6},
        }

        self.phase_equilibrium_list = {
            "e1": ["c1", ("p1", "p2")],
            "e2": ["c2", ("p1", "p2")],
        }

        # Add inherent reactions for use when needed
        self.inherent_reaction_idx = Set(initialize=["i1", "i2"])
        self.inherent_reaction_stoichiometry = {
            ("i1", "p1", "c1"): 1,
            ("i1", "p1", "c2"): 1,
            ("i1", "p2", "c1"): 1,
            ("i1", "p2", "c2"): 1,
            ("i2", "p1", "c1"): 1,
            ("i2", "p1", "c2"): 1,
            ("i2", "p2", "c1"): 1,
            ("i2", "p2", "c2"): 1,
        }

        # Attribute to switch flow basis for testing
        self.basis_switch = 1
        self.default_balance_switch = 1

        self._state_block_class = StateBlockForTesting

        self.set_default_scaling("flow_vol", 100)
        self.set_default_scaling("flow_mol", 101)
        self.set_default_scaling("flow_mol_phase_comp", 102)
        self.set_default_scaling("test_var", 103)
        self.set_default_scaling("pressure", 104)
        self.set_default_scaling("temperature", 105)
        self.set_default_scaling("enth_mol", 106)
        self.set_default_scaling("gibbs_mol_phase_comp", 107)
        self.set_default_scaling("entr_mol", 108)
        self.set_default_scaling("mole_frac_phase_comp", 109)
        self.set_default_scaling("enthalpy_flow", 110)
        self.set_default_scaling("energy_dens", 111)
        self.set_default_scaling("material_flow_mol", 112)
        self.set_default_scaling("material_dens_mol", 113)
        self.set_default_scaling("material_flow_mass", 114)
        self.set_default_scaling("material_dens_mass", 115)
        self.set_default_scaling("material_flow_dimensionless", 116)
        self.set_default_scaling("material_dens_dimensionless", 117)

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "enth_mol": {"method": None},
            }
        )
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


class SBlockBase(StateBlock):
    def initialize(
        blk,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=False,
        state_args=None,
    ):
        for k in blk.values():
            k.init_test = True
            k.hold_state = hold_state

    def release_state(blk, flags=None, outlvl=idaeslog.NOTSET):
        for k in blk.values():
            k.hold_state = not k.hold_state


@declare_process_block_class("StateBlockForTesting", block_class=SBlockBase)
class StateTestBlockData(StateBlockData):
    CONFIG = ConfigBlock(implicit=True)

    def build(self):
        super(StateTestBlockData, self).build()

        self.flow_vol = Var(initialize=20, units=units.m**3 / units.s)
        self.flow_mol = Var(
            initialize=14,
            units=units.mol / units.s,
        )
        self.flow_mol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=2,
            units=units.mol / units.s,
        )
        self.test_var = Var(initialize=1)
        self.enthalpy_flow = Var(initialize=1, units=units.J / units.s)
        self.energy_dens = Var(initialize=1, units=units.J / units.m**3)
        self.material_flow_mol = Var(initialize=1, units=units.mol / units.s)
        self.material_dens_mol = Var(initialize=1, units=units.mol / units.m**3)
        self.material_flow_mass = Var(initialize=1, units=units.kg / units.s)
        self.material_dens_mass = Var(initialize=1, units=units.kg / units.m**3)
        self.material_flow_dimensionless = Var(initialize=1, units=units.dimensionless)
        self.material_dens_dimensionless = Var(initialize=1, units=units.dimensionless)
        self.pressure = Var(initialize=1e5, units=units.Pa)
        self.temperature = Var(initialize=300, units=units.K)

        self.cp_mol = Var(initialize=100, units=units.J / units.mol / units.K)
        self.enth_mol = Var(initialize=10000, units=units.J / units.mol)

        self.gibbs_mol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=50,
            units=units.J / units.mol,
        )
        self.entr_mol = Var(initialize=1000, units=units.J / units.mol / units.K)

        self.mole_frac_phase_comp = Var(
            self.params.phase_list, self.params.component_list, initialize=0.5
        )

    def get_material_flow_terms(b, p, j):
        if b.config.parameters.basis_switch == 2:
            return b.material_flow_mass
        elif b.config.parameters.basis_switch == 1:
            return b.material_flow_mol
        else:
            return b.material_flow_dimensionless

    def get_material_density_terms(b, p, j):
        if b.config.parameters.basis_switch == 2:
            return b.material_dens_mass
        elif b.config.parameters.basis_switch == 1:
            return b.material_dens_mol
        else:
            return b.material_dens_dimensionless

    def get_enthalpy_flow_terms(b, p):
        return b.enthalpy_flow

    def get_energy_density_terms(b, p):
        return b.energy_dens

    def model_check(self):
        self.check = True

    def get_material_flow_basis(b):
        if b.config.parameters.basis_switch == 1:
            return MaterialFlowBasis.molar
        elif b.config.parameters.basis_switch == 2:
            return MaterialFlowBasis.mass
        else:
            return MaterialFlowBasis.other

    def default_material_balance_type(self):
        if self.params.default_balance_switch == 1:
            return MaterialBalanceType.componentPhase
        else:
            raise NotImplementedError

    def default_energy_balance_type(self):
        if self.params.default_balance_switch == 1:
            return EnergyBalanceType.enthalpyTotal
        else:
            raise NotImplementedError

    def define_state_vars(self):
        return {
            "component_flow_phase": self.flow_mol_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(_ReactionParameterBlock, self).build()

        self.rate_reaction_idx = Set(initialize=["r1", "r2"])
        self.equilibrium_reaction_idx = Set(initialize=["e1", "e2"])

        self.rate_reaction_stoichiometry = {
            ("r1", "p1", "c1"): 1,
            ("r1", "p1", "c2"): 1,
            ("r1", "p2", "c1"): 1,
            ("r1", "p2", "c2"): 1,
            ("r2", "p1", "c1"): 1,
            ("r2", "p1", "c2"): 1,
            ("r2", "p2", "c1"): 1,
            ("r2", "p2", "c2"): 1,
        }
        self.equilibrium_reaction_stoichiometry = {
            ("e1", "p1", "c1"): 1,
            ("e1", "p1", "c2"): 1,
            ("e1", "p2", "c1"): 1,
            ("e1", "p2", "c2"): 1,
            ("e2", "p1", "c1"): 1,
            ("e2", "p1", "c2"): 1,
            ("e2", "p2", "c1"): 1,
            ("e2", "p2", "c2"): 1,
        }

        self._reaction_block_class = ReactionBlock

        # Attribute to switch flow basis for testing
        self.basis_switch = 1

        self.set_default_scaling("reaction_rate", 101, "r1")
        self.set_default_scaling("reaction_rate", 102, "r2")

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )

    @classmethod
    def get_required_properties(cls):
        return {}


class RBlockBase(ReactionBlockBase):
    def initialize(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, state_vars_fixed=False
    ):
        for k in blk.values():
            k.init_test = True


@declare_process_block_class("ReactionBlock", block_class=RBlockBase)
class ReactionBlockData(ReactionBlockDataBase):
    CONFIG = ConfigBlock(implicit=True)

    def build(self):
        super(ReactionBlockData, self).build()

        self.reaction_rate = Var(["r1", "r2"], units=units.mol / units.m**3 / units.s)

        self.dh_rxn = {
            "r1": 10 * units.J / units.mol,
            "r2": 20 * units.J / units.mol,
            "e1": 30 * units.J / units.mol,
            "e2": 40 * units.J / units.mol,
        }

    def model_check(self):
        self.check = True

    def get_reaction_rate_basis(b):
        if b.config.parameters.basis_switch == 1:
            return MaterialFlowBasis.molar
        elif b.config.parameters.basis_switch == 2:
            return MaterialFlowBasis.mass
        else:
            return MaterialFlowBasis.other


def _enable_scip_solver_for_testing(
    name: str = "scip",
) -> Union[Callable[[], None], None]:
    ampl_module_scip, is_available = attempt_import("ampl_module_scip")
    if not is_available:
        _log.warning(
            "ampl_module_scip must be installed to enable SCIP solver for testing"
        )
        return None

    new_path_entry = ampl_module_scip.bin_dir
    # TODO prepending new_path_entry to PATH means that the "testing" SCIP will "shadow"
    # existing executable directories already on PATH
    # this behavior would match the (implied) semantics of this function, i.e.
    # "enable SCIP solver used for testing (at the expense of other non-testing SCIP executables)"
    os.environ["PATH"] = os.pathsep.join([new_path_entry, os.environ["PATH"]])
    Executable.rehash()

    def remove_from_path():
        path_as_list = os.environ["PATH"].split(os.pathsep)

        try:
            path_as_list.remove(new_path_entry)
        except ValueError:
            # not in PATH
            pass
        else:
            os.environ["PATH"] = os.pathsep.join(path_as_list)
            Executable.rehash()

    return remove_from_path
